/*
 * $Id: arch-dunix-distrib.h,v 1.1.2.5 1997/11/25 05:03:25 randall Exp $
 */
/*
 * $Log: arch-dunix-distrib.h,v $
 * Revision 1.1.2.5  1997/11/25 05:03:25  randall
 * Fixed timers.
 *
 * Revision 1.1.2.4  1997/11/10 22:27:52  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.1.2.3  1997/10/22 02:17:12  randall
 * More iterations of getting things working.
 *
 * Revision 1.1.2.2  1997/10/22 01:25:18  randall
 * Got DEC version working again.
 *
 * Revision 1.1.2.1  1997/10/14 20:25:27  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.1  1997/09/22 20:27:21  randall
 * Got process version of Cilk working.  Shared globals still aren't working!!!
 *
 * Revision 1.1.1.1  1997/02/14 23:07:12  strumpen
 * Create Cilk Repository
 *
 * Revision 1.1.1.1  1997/02/14 22:04:29  strumpen
 * Create cilk repository
 *
 * Revision 1.1  1997/01/16 14:23:05  athena
 * Added support for linux
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#define __C_ASM_H
#include <unistd.h>
void usleep(useconds_t useconds);  /* This doesn't seem to be in unistd.h like it is supposed to be... */
#include <signal.h>


#define CACHE_LINE 64

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

#if (!CILK_DUNIX_DISTRIB)
#error CILK_DUNIX_DISTRIB not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

#define CILK_USE_MEMORY_LOCKS 1

/* a lock variable */
#if CILK_USE_MEMORY_LOCKS
typedef volatile int Cilk_mutex;
#else
#error OS locks not implemented yet
#endif

/***********************************************************\
 * Machine size
\***********************************************************/

static inline int Cilk_partition_size(void)
{
     return sysconf(_SC_NPROCESSORS_ONLN);
}

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     { usleep(10000); } /* ??? */

/***********************************************************\
 * Various types of membars
\***********************************************************/

/*
 * Generic memory barrier
 */
#define CILK_SYNC asm volatile ("mb":::"memory")
/*
 * Write-only memory barrier
 */
#define CILK_W_SYNC asm volatile ("wmb":::"memory")

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
  CILK_SYNC;
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
  CILK_W_SYNC;
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
    CILK_SYNC;
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/times.h>
#include <sys/time.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long Cilk_time;

Cilk_time Cilk_get_time(void); /* implemented in arch-dunix-distrib.c */

static inline double Cilk_time_to_sec(Cilk_time t)
{
  /* any way to get clock frequency at run-time? */
     return (double) t / 466e6;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
  struct timeval t;
  gettimeofday(&t, NULL);
  return 1000000LL * t.tv_sec + t.tv_usec;
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return t / 1000000.0;
}

static inline void Cilk_init_timers()
{
  /* calibrate clock? */
}

/***********************************************************\
 * Fast user locks (spin locks)
\***********************************************************/

typedef volatile int Cilk_lockvar;

static inline void Cilk_lock_init(Cilk_lockvar *v)
{
  *v = 0;
  CILK_W_SYNC;
}

static inline void Cilk_lock(Cilk_lockvar *v)
{
  int tmp;
  
  do {
    do {
      asm volatile ("ldl_l %0, (%1)":"=r" (tmp):"r"(v));
    } while (tmp != 0);
    
    tmp = 1;
    asm volatile ("stl_c %0, (%1)":"=r" (tmp):"r"(v), "0"(tmp));
  } while (tmp == 0);
  CILK_SYNC;
}

static inline void Cilk_unlock(Cilk_lockvar *v)
{
  /* Note: we can't use CILK_W_SYNC here unless conditions on
   * page 5-9,10 in the Alpha Architecture Handbook are satisfied.
   * http://ftp.digital.com/pub/Digital/info/semiconductor/literature/alphahb2.pdf
   */
  CILK_SYNC;
  *v = 0;
}

/***********************************************************\
 * Typedefs for messaging library
\***********************************************************/

#define MAXMACHINES 100
#define MAX_SEND_QUEUE 1024
#define MAX_MISSING 128

/* header for packets in transit */
typedef struct
{
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int first_seq_no;          /* first seq no of this message */
  int sender;                /* message sender               */
  int seq_no;                /* sequence number from sender  */
  int ack_seq_no;            /* piggy-backed acknowledgement */
} TransitHeader;

/* header for messages waiting to be sent */
typedef struct SendMessage
{
  struct SendMessage *next;
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int buf_size;              /* # of bytes to free when done */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int dest;                  /* destination machine          */
  void *data;                /* data of message (body)       */
} SendMessage;

/* header for packets being reconstructed at receiver */
typedef struct RecvMessage
{
  struct RecvMessage *next;
  int size;                  /* # of bytes in message body   */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int chunks_so_far;         /* # of chunks received         */
  void *data;                /* data of message (body)       */
} RecvMessage;

/* data type for keeping track of which packets need to be sent */
typedef struct
{
  int seq_no;
  SendMessage *msg;
} SendReq;

/* data type for keeping track of which packets we still need from sender */
typedef struct
{
  int sender;                /* message sender               */
  int seq_no;                /* sequence # of missing chunk  */
  Cilk_time gone_missing;    /* time we noticed absence      */
} MissingPacket;

/***********************************************************\
 * Typedefs for cache
\***********************************************************/

/*
 * This number should be the size of the largest virtually indexed
 * cache in the system, divided by its associativity and the
 * system page size.  For example, on SuperSparc processors, there
 * is a 16k virtually indexed direct-mapped L1 cache, and a page
 * size of 8k.  This gives a color value of (16k / 1 / 8k) = 2.
 */
/* For digital machines, there is an 8k virtually indexed direct
 * mapped cache, with a page size of 8k ==> VCOLORS=1.
 */
#define VCOLORS 1
