#include <shmem-distrib.h>
#include <malloc.h>

#define TRACE(x,y)

/***********************************************************\
 * Bit vectors
\***********************************************************/

typedef unsigned int BitVector;

static inline int get_bit(BitVector *b, int n)
{
  CILK_ASSERT(n>=0 && n<32);
  return (*b) & (1<<n);
}

static inline void set_bit(BitVector *b, int n)
{
  CILK_ASSERT(n>=0 && n<32);
  (*b) |= (1<<n);
}

static inline void clear_bit(BitVector *b, int n)
{
  CILK_ASSERT(n>=0 && n<32);
  (*b) &= ~(1<<n);
}

static inline int any_bits(BitVector *b)
{
  return (*b);
}

/*
 * Clear any bits in b which are not set in mask.
 */
static inline void mask_bits(BitVector *b, BitVector *mask)
{
  (*b) &= (*mask);
}

#define EMPTY_BIT_VECTOR 0

/***********************************************************\
 * Cache structure and global variables
\***********************************************************/

typedef struct CachePage
{
  struct CachePage *prev, *next;  /* doubly-linked hash list             */
  struct CachePage *forw, *back;  /* doubly-linked state list            */
  void *page;             /* virtual address of page                     */
  int stack_id;           /* ID of branch in cactus stack                */

  int pins;               /* # of processes pinning page in cache        */
  void *twin;             /* copy at page-in time (NULL if read-only)    */
  void *working;          /* current copy                                */
  BitVector read;         /* vector of processes mapping page readable   */
  BitVector write;        /* vector of processes mapping page writeable  */
  BitVector dirty;        /* vector of processes for which page is dirty */
  int op_in_progress;     /* set if awaiting a backing store op          */
  int timestamp;          /* timestamp of last fetch from bstore         */
  BitVector uncleanable;  /* vector of processes which can't be made
			   * clean when reconcile reply is received.
			   * Valid only when op_in_prog == RECONCILE_IN_PROG */
} CachePage;

/* <page> and <stack_id> cannot be changed while <pins> > 0.  I.e., a
 * CachePage structure may not be reallocated to another page while <pins> > 0.
 *
 */

#define FETCH_IN_PROGRESS 1
#define RECONCILE_IN_PROGRESS 2
#define TWIN_IN_PROGRESS 3

DECLARE_MACHINE_VARIABLE(int, pagesize, );
DECLARE_MACHINE_VARIABLE(static int, cache_pages, );

DECLARE_MACHINE_VARIABLE(static int *, page_free_array, );
DECLARE_MACHINE_VARIABLE(static int, page_free_list, [VCOLORS]);

DECLARE_MACHINE_VARIABLE(static struct CachePage *, header_free_list, );

DECLARE_MACHINE_VARIABLE(static int, cache_hashsize, );
DECLARE_MACHINE_VARIABLE(static struct CachePage *, hasht, );

/* LRU list of pages currently in cache.
 */
DECLARE_MACHINE_VARIABLE(static struct CachePage *, mapped_pages, );

/* unmapped_pages is not currently being used !!! */
DECLARE_MACHINE_VARIABLE(static struct CachePage *, unmapped_pages, );

DECLARE_MACHINE_VARIABLE(static void *, cache_mem_image, );

DECLARE_MACHINE_VARIABLE(static int, latest_timestamp, );
DECLARE_MACHINE_VARIABLE(static int, local_timestamp, [32]);

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, cache_lock, );

/*
 * Invariants: Any time cache_lock is not held,
 * - a page is mapped readable iff the read bit is set.
 * - a page is mapped writeable iff the write bit is set.
 * - any page with a write bit set also has the corresponding read bit set.
 * - any page with a write bit set also has the corresponding dirty bit set.
 * - twin != NULL iff there is a dirty bit set
 * - there are no pages in the cache with a timestamp greater than
 *   latest_timestamp.
 * - a computation is permitted to access (read) any page in the cache
 *   which has a timestamp greater than its local_timestamp
 * - local_timestamp is even.
 * - latest_timestamp and the timestamps of pages are odd.
 * - if op_in_progress is not zero, then pins is not zero.
 * - there is only one backing store request in flight at a time.
 * - a page is in mapped_pages iff it has a read or dirty bit set - otherwise
 *   it is in unmapped_pages.
 *
 * Facts
 * - at release time, process p only has to reconcile pages which
 *   have their dirty bit for p set.
 * - at acquire time, process p only has to unmap pages which
 *   have their read bit for p set.
 *
 *
 * Notes
 * - A page fetch must be given the timestamp of when it is
 *   requested, not when the data arrives.  Therefore, timestamps are
 *   passed around in FetchRequest/FetchReply structures so that
 *   the timestamp can be set at the right time.
 */

/***********************************************************\
 * Utility routines
\***********************************************************/

static inline int prime(unsigned x)
{
  unsigned i;
  
  for(i=2; i*i <= x; i++)
    if (x%i == 0)
      return 0;
  return 1;
}
static inline unsigned next_prime(unsigned x)
{
  for(;1;x++)
    if (prime(x))
      return x;
}

/*
 * Note: virtual page color is absolute, not relative to
 * cache_mem_image.  In other words, page 0 in the cache
 * is not necessarily color 0.
 */
static inline int page_color(void *p)
{
  return (size_t)p / USE_MACHINE(pagesize) % VCOLORS;
}

/*
 * Hash which computes the machine on which the backing store for
 * the given page is stored.
 */
static inline int machine_hash(void *page, int stack_id)
{
  /* simple striped hash for now */
  return ((size_t)page / USE_MACHINE(pagesize)) % USE_MACHINE(Machines);
}


/*
 * List operations
 */
static inline void splice_out_hash(CachePage *c)
{
  CachePage *next, *prev;
  prev = c->prev;
  next = c->next;
  prev->next = next;
  next->prev = prev;
  WHEN_CILK_DEBUG(c->prev = NULL);
  WHEN_CILK_DEBUG(c->next = NULL);
}

static inline void splice_in_hash(CachePage *c, CachePage *prev)
{
  CachePage *next = prev->next;
  prev->next = c;
  next->prev = c;
  c->prev = prev;
  c->next = next;
}

static inline void splice_out_lru(CachePage *c)
{
  CachePage *back, *forw;
  back = c->back;
  forw = c->forw;
  back->forw = forw;
  forw->back = back;
  WHEN_CILK_DEBUG(c->back = NULL);
  WHEN_CILK_DEBUG(c->forw = NULL);
}

static inline void splice_in_lru(CachePage *c, CachePage *forw)
{
  CachePage *back = forw->back;
  back->forw = c;
  forw->back = c;
  c->back = back;
  c->forw = forw;
}

/*
 * Call this when a CachePage's status has changed to be less mappable.
 */
static inline void update_unmap(CachePage *c)
{
  if (!any_bits(&c->read) && !any_bits(&c->dirty))
    {
      CILK_ASSERT(!any_bits(&c->write));
      splice_out_lru(c);
      splice_in_lru(c, USE_MACHINE(unmapped_pages));
    }
}
/*
 * Call this when a CachePage's status has changed to be more mappable,
 * i.e. if a read or dirty bit is set.
 */
static inline void update_map(CachePage *c)
{
  /* just move page to head of "lru" list */
  splice_out_lru(c);
  splice_in_lru(c, USE_MACHINE(mapped_pages));
}

/***********************************************************\
 * Cactus Stack - doesn't exist anymore!
\***********************************************************/

static inline int find_stack_id(void *page)
{
  return 42;
}

/***********************************************************\
 * Cache
\***********************************************************/

void Cilk_cache_init(const Options *options)
{
  size_t size = options->cache;
  CachePage *cache_headers;
  int i;
  
  TRACE(TRACE_CACHE, ("setting up cache for machine %d\n", Machine));
  
  /*
   * Set up page cache in memory, link pages into free list.
   */
  size = (size + USE_MACHINE(pagesize) - 1) & ~(USE_MACHINE(pagesize)-1);
  USE_MACHINE(cache_mem_image) = Cilk_setup_cache(size, options);
  USE_MACHINE(cache_pages) = size / USE_MACHINE(pagesize);
  
  /*
   * Allocate and initialize free list structure
   */
  USE_MACHINE(page_free_array) = Cilk_malloc_machine_fixed(USE_MACHINE(cache_pages) * sizeof(int));
  for(i=0; i<VCOLORS; i++)
    USE_MACHINE(page_free_list)[i] = -1;
  for(i=0; i<USE_MACHINE(cache_pages); i++)
    {
      int color = page_color(USE_MACHINE(cache_mem_image) + i*USE_MACHINE(pagesize));
      USE_MACHINE(page_free_array)[i] = USE_MACHINE(page_free_list)[color];
      USE_MACHINE(page_free_list)[color] = i;
    }
  
  /*
   * Set up lru lists.
   */
  USE_MACHINE(mapped_pages) = Cilk_malloc_machine_fixed(sizeof(CachePage));
  USE_MACHINE(unmapped_pages) = Cilk_malloc_machine_fixed(sizeof(CachePage));
  USE_MACHINE(mapped_pages)->forw = USE_MACHINE(mapped_pages);
  USE_MACHINE(mapped_pages)->back = USE_MACHINE(mapped_pages);
  USE_MACHINE(unmapped_pages)->forw = USE_MACHINE(unmapped_pages);
  USE_MACHINE(unmapped_pages)->back = USE_MACHINE(unmapped_pages);
  
  /*
   * Allocate headers for pages.
   */
  cache_headers = (CachePage*)Cilk_malloc_machine_fixed(USE_MACHINE(cache_pages) * sizeof(CachePage));
  CILK_CHECK(cache_headers != NULL,
	     ("no memory for cache headers!\n"));
  for(i=0; i < USE_MACHINE(cache_pages) - 1; i++)
    cache_headers[i].next = &cache_headers[i+1];
  cache_headers[i].next = NULL;
  USE_MACHINE(header_free_list) = cache_headers;
  
  /*
   * Allocate hash table.
   */
  USE_MACHINE(cache_hashsize) = next_prime(2*USE_MACHINE(cache_pages));
  
  USE_MACHINE(hasht) = (CachePage*) Cilk_malloc_machine_fixed(sizeof(CachePage) * USE_MACHINE(cache_hashsize));
  CILK_CHECK(USE_MACHINE(hasht) != NULL,
	     ("couldn't allocate hash table!\n"));
  for(i=0; i<USE_MACHINE(cache_hashsize); i++)
    USE_MACHINE(hasht)[i].prev = USE_MACHINE(hasht)[i].next = &USE_MACHINE(hasht)[i];
  
  USE_MACHINE(latest_timestamp) = 3;
  for(i=0; i<32; i++)
    USE_MACHINE(local_timestamp)[i] = 2;
  
  Cilk_mutex_init(&USE_MACHINE(cache_lock));
}

static inline int cache_hash(void *page, int stack_id)
{
  return ((size_t)page / USE_MACHINE(pagesize)) % USE_MACHINE(cache_hashsize);
}

static inline CachePage *find_page(void *page, int stack_id)
{
  CachePage *t = &USE_MACHINE(hasht)[cache_hash(page, stack_id)];
  CachePage *c;
  for(c = t->next; c!=t; c=c->next)
    if (c->page == page && c->stack_id == stack_id)
      {
#if CILK_DEBUG
	CachePage *c2;
	for(c2 = c->next; c2!=t; c2=c2->next)
	  if (c2->page == page && c2->stack_id == stack_id)
	    Cilk_die("double page!\n");
#endif
	return c;
      }
  return NULL;
}

static inline void free_cache_page(void *p)
{
  int id = (p - USE_MACHINE(cache_mem_image)) / USE_MACHINE(pagesize);
  int color = page_color(p);
  
  USE_MACHINE(page_free_array)[id] = USE_MACHINE(page_free_list)[color];
  USE_MACHINE(page_free_list)[color] = id;
}

static void drop_page(CachePage *c)
{
  void *p;
  
  CILK_ASSERT(!c->pins);
  CILK_ASSERT(!c->op_in_progress);
  CILK_ASSERT(!any_bits(&c->read));
  CILK_ASSERT(!any_bits(&c->write));
  CILK_ASSERT(!any_bits(&c->dirty));
  
  /* splice page out of hash table, lru list */
  splice_out_hash(c);
  splice_out_lru(c);
  
  /* free working page */
  p = c->working;
  free_cache_page(p);
  
  /* free twin page */
  CILK_ASSERT(c->twin == NULL);  /* guaranteed by invariant */
  
  /* free header */
  c->next = USE_MACHINE(header_free_list);
  USE_MACHINE(header_free_list) = c;
}

/***********************************************************\
 * Page eviction
\***********************************************************/

static void reconcile_page(CachePage *c);

DECLARE_MACHINE_VARIABLE(static unsigned, eviction_color, );
DECLARE_MACHINE_VARIABLE(static volatile int, eviction_done, [32]);

#define EVICT_COUNT 10

/*
 * Unmap EVICT_COUNT pages from the LRU list, except possibly
 * pages with an event in progress.
 */
/* This code is called from within an asynchronous interrupt, so we
 * can't do very much at all.  NO lock acquires, and no calls to
 * timer/stats routines.  Preferrably, no writes at all except for
 * eviction_done.
 */
void Cilk_cache_evict_pages(void)
{
  int count = 0;
  CachePage *c;
  
  /*
   * Unmap first EVICT_COUNT eligible pages, if we have them mapped.
   */
  for(c = USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages) && count < EVICT_COUNT; c=c->forw)
    {
      if (!c->pins &&
	  (page_color(c->page) == USE_MACHINE(eviction_color) ||
	   (any_bits(&c->dirty) &&
	    page_color(c->page) == (USE_MACHINE(eviction_color)+VCOLORS-1) % VCOLORS)))
	{
	  if (get_bit(&c->read, Process))
	    Cilk_map_cache_page(c->page, NULL, CILK_MAP_FAULTING | CILK_MAP_NOSTAT);
	  count++;
	}
    }
  
  /* signal that this process is done unmapping */
  USE_MACHINE(eviction_done)[USE_PRIVATE(Process)] = 1;
}

/*
 * Make more space in the cache.
 * May release and reacquire cache_lock.
 */
static void make_more_space(unsigned color)
{
  CachePage *c, *next;
  int count;
  int i;
  /* Cilk_dprintf("make more space!\n"); */
  
loop:
  /*
   * Drop or reconcile a page in pages, if it has no readers or writers.
   */
  for(c = USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages); c=c->forw)
    {
      if (!c->pins &&
	  !any_bits(&c->read) &&
	  !any_bits(&c->dirty) &&
	  page_color(c->page) == color)
	{
	  /* Cilk_dprintf("page dropped\n"); */
	  drop_page(c);
	  return;
	}
      if (!c->pins &&
	  !any_bits(&c->read) &&
	  any_bits(&c->dirty) &&
	  (page_color(c->page) == color ||
	   page_color(c->page) == (color+VCOLORS-1)%VCOLORS))
	{
	  c->pins++;
	  reconcile_page(c);
	  c->pins--;
	  goto loop;  /* try again */
	}
    }
  
  /*
   * Signal all other processes to unmap the first EVICT_COUNT eligible
   * pages in the mapped_pages list.
   * Note that we have the cache lock at this point.
   */
  /* Cilk_dprintf("doing the eviction thing...\n"); */
  
  /* setup eviction */
  USE_MACHINE(eviction_color) = color;
  
  Cilk_enter_state(STATE_USER7);
  
  /* send eviction request */
  for(i=0; i<USE_MACHINE(Processes); i++)
    {
      USE_MACHINE(eviction_done)[i] = 0;
      if (i != Process)
	Cilk_send_eviction_notice(i);
    }
  Cilk_cache_evict_pages();          /* no need to signal ourselves */
  
  /* wait for everyone to evict pages */
  do
    {
      for(i=0; i<USE_MACHINE(Processes); i++)
	if (!USE_MACHINE(eviction_done)[i])
	  break;
    }
  while(i != USE_MACHINE(Processes));
  
  Cilk_exit_state(STATE_USER7);
  
  /* Cilk_dprintf("received response from everyone\n"); */
  /*
   * Now, update pages to reflect the fact that they were unmapped
   */
  count = 0;
  for(c = USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages) && count < EVICT_COUNT; c=next)
    {
      next = c->forw;
      CILK_ASSERT(c != NULL);
      if (!c->pins &&
	  (page_color(c->page) == USE_MACHINE(eviction_color) ||
	   (any_bits(&c->dirty) &&
	    page_color(c->page) == (USE_MACHINE(eviction_color)+VCOLORS-1) % VCOLORS)))
	{
	  /* Cilk_dprintf("unmapping %p\n", c); */
	  CILK_ASSERT(!c->op_in_progress);
	  c->read = EMPTY_BIT_VECTOR;
	  c->write = EMPTY_BIT_VECTOR;
	  /* update_unmap(c); */
	  count++;
	}
    }
  /* Cilk_dprintf("looping\n"); */
  goto loop;
}

/* may release and reacquire cache_lock */
static void *get_cache_page(int color)
{
  int id;
  void *p;
  
  while((id = USE_MACHINE(page_free_list)[color]) == -1)
    {
      make_more_space(color);
    }
  USE_MACHINE(page_free_list)[color] = USE_MACHINE(page_free_array)[id];
  WHEN_CILK_DEBUG(USE_MACHINE(page_free_array)[id] = -1);
  
  p = USE_MACHINE(cache_mem_image) + id * USE_MACHINE(pagesize);
  CILK_ASSERT(page_color(p) == color);
  
  return p;
}

/***********************************************************\
 * Reconciling
\***********************************************************/

void Cilk_cache_reconcile_reply(ReconcileReply *reply)
{
  CachePage *c = reply->c;
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  CILK_ASSERT(c->op_in_progress == RECONCILE_IN_PROGRESS);
  CILK_ASSERT(c->pins > 0);
  CILK_ASSERT(any_bits(&c->dirty));
  
  /*
   * Anyone who hasn't mapped this page writeable since the
   * reconcile data was computed is now clean.
   */
  mask_bits(&c->dirty, &c->uncleanable);
  
  c->op_in_progress = 0;
  if (!any_bits(&c->dirty))
    {
      CILK_ASSERT(!any_bits(&c->write));
      free_cache_page(c->twin);
      WHEN_CILK_DEBUG(c->twin = NULL);
      /* update_unmap(c); ???!!!*/
    }
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
}

/*
 * Requires:  Holding cache lock on entry.
 * May release and reacquire cache lock.
 */
static void reconcile_page(CachePage *c)
{
  int pn = -1;
  ReconcileRequest *req;
  int msg_size, size;
  
  CILK_ASSERT(any_bits(&c->dirty));
  CILK_ASSERT(c->twin != NULL);
  CILK_ASSERT(c->pins > 0);
  CILK_ASSERT(!c->op_in_progress);
  c->op_in_progress = RECONCILE_IN_PROGRESS;
  
  /* save vector of uncleanable processes */
  c->uncleanable = c->write;
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
  
  /*
   * compute diff
   */
  pn = machine_hash(c->page, c->stack_id);
  msg_size = sizeof(ReconcileRequest) + USE_MACHINE(pagesize) / sizeof(int) * sizeof(RecEntry);
  req = Cilk_alloc_msg_buf(msg_size, pn);
  
  size = req->size = compute_diff(c->working, c->twin, req->diff);
  
  /* if nothing is dirty, nothing to do! */
  if (size == 0)
    {
      ReconcileReply reply;
      
      Cilk_free_msg_buf(req, msg_size, pn);
      
      reply.c = c;
      Cilk_cache_reconcile_reply(&reply);
      Cilk_event(EVENT_RECONCILE_NULL);
      TRACE(TRACE_RECONCILE, ("null reconcile of page %p(%d)\n", c->page, c->stack_id));
    }
  else
    {
      /* fill in request structure */
      req->page = c->page;
      req->stack_id = c->stack_id;
      req->machine = Machine;
      req->c = c;
      
      /* send request */
      TRACE(TRACE_RECONCILE, ("reconciling page %p(%d) to %d\n", c->page, c->stack_id, pn));
      
      Cilk_rpc_with_size(pn, Cilk_bstore_reconcile_req, req, msg_size, sizeof(ReconcileRequest) + size * sizeof(RecEntry));
      
      Cilk_event(EVENT_RECONCILE);
      
      /*
       * Wait for reconcile to complete
       */
      Cilk_enter_state(STATE_WAITING_FOR_RECONCILE);
      while(c->op_in_progress == RECONCILE_IN_PROGRESS)
	Cilk_poll();
      Cilk_exit_state(STATE_WAITING_FOR_RECONCILE);
    }
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
}

void Cilk_cache_fetch_reply(FetchReply *reply)
{
  CachePage *c = reply->c;
  int zero = reply->zero;
  int timestamp = reply->timestamp;
  void *data = reply->data;
  
  /* Cilk_dprintf("got page %p(%d)\n", c->page, c->stack_id); */
  
  if (zero)
    Cilk_event(EVENT_FETCH_ZERO);
  else
    Cilk_event(EVENT_FETCH);
  
  CILK_ASSERT(c->op_in_progress == FETCH_IN_PROGRESS);
  CILK_ASSERT(c->pins > 0);
  
  if (any_bits(&c->dirty))
    {
      /* the page is dirty - diff data with twin page,
       * apply resulting diff to twin and working pages
       */
      CILK_ASSERT(c->twin != NULL);
      TRACE(TRACE_CACHE, ("Received page %p(%d), applying to dirty page\n", c->page, c->stack_id));
      if (zero)
	apply_zero_data(c->working, c->twin);
      else
	apply_data(data, c->working, c->twin);
    }
  else
    {
      /* read-only page - just overwrite with data */
      if (zero)
	{
	  TRACE(TRACE_CACHE, ("Received zero page %p(%d)\n", c->page, c->stack_id));
	  zero_page(c->working);
	}
      else
	{
	  TRACE(TRACE_CACHE, ("Received page %p(%d)\n", c->page, c->stack_id));
	  copy_page(data, c->working);
	}
    }
  
  Cilk_membar_StoreStore();
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  CILK_ASSERT(c->timestamp < timestamp);
  c->timestamp = timestamp;
  
  CILK_ASSERT(c->op_in_progress == FETCH_IN_PROGRESS);
  c->op_in_progress = 0;
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
}

/*
 * Wait until any other operation on this page has completed.
 * Page lock should be held upon entry, and is held upon exit.
 * Lock may be released and reacquired.  The pointer <c> is
 * still valid upon return, but any values c points to must be
 * reloaded.
 */
static void wait_for_turn(CachePage *c)
{
  CILK_ASSERT(c->pins > 0);
  while(c->op_in_progress)
    {
      Cilk_mutex_signal(&USE_MACHINE(cache_lock));
      while(c->op_in_progress)
	Cilk_poll();
      Cilk_mutex_wait(&USE_MACHINE(cache_lock));
    }
}

/*
 * Lock is held upon entry.  Lock is still held on exit.
 */
static void write_fault(CachePage *c)
{
  void *page = c->page;
  void *work = c->working;
  
  /* Write fault - map the page read-write and return */
  CILK_ASSERT(get_bit(&c->read, Process));
  CILK_ASSERT(!get_bit(&c->write, Process));
  CILK_ASSERT(c->pins > 0);
  
  /* get a twin page in case we need it */
  if (!any_bits(&c->dirty))
    {
      void *twin;
      
      twin = get_cache_page((page_color(page)+1) % VCOLORS);
      CILK_ASSERT(twin != NULL);
      
      wait_for_turn(c);
      
      if (get_bit(&c->write, Process))
	{
	  /* write fault handled by someone else */
	  free_cache_page(twin);
	  return;
	}
      if (!get_bit(&c->read, Process))
	{
	  /* page unmapped in the meantime */
	  free_cache_page(twin);
	  return;        /* fault will be retried, this time as a read fault */
	}
      if (any_bits(&c->dirty))
	{
	  /* someone else made a twin */
	  free_cache_page(twin);
	  CILK_ASSERT(c->twin != NULL);
	}
      else
	{
	  CILK_ASSERT(!c->op_in_progress);
	  c->op_in_progress = TWIN_IN_PROGRESS;
	  c->twin = twin;
	  
	  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
	  
	  make_twin(work, twin);
	  
	  TRACE(TRACE_CACHE, ("made twin for page %p(%d)\n", page, c->stack_id));
	  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
	  CILK_ASSERT(c->op_in_progress == TWIN_IN_PROGRESS);
	  c->op_in_progress = 0;
	}
    }
  
  set_bit(&c->write, Process);
  set_bit(&c->dirty, Process);
  set_bit(&c->uncleanable, Process);
  /* don't need to update_map, because c->read != 0 already */
  
  /* put map call outside cache lock because it is slow */
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
  Cilk_map_cache_page(page, work, CILK_MAP_READWRITE);
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  TRACE(TRACE_CACHE, ("mapped page %p(%d) writeable\n", page, c->stack_id));
}

/*
 * Page lock is held upon entry.  Lock is still held at exit.
 */
static void read_fault(CachePage *c)
{
  void *page = c->page;
  int stack_id = c->stack_id;
  int try = 0;
  
  CILK_ASSERT(!get_bit(&c->read, Process));
  CILK_ASSERT(!get_bit(&c->write, Process));
  CILK_ASSERT(c->pins > 0);
  
again:
  /* check if we are allowed access to page */
  if (c->timestamp < USE_MACHINE(local_timestamp)[Process])
    {
      /* We are not allowed access -- send a fetch request to get
       * a more up-to-date copy of the page.
       */
      FetchRequest *req;
      int pn;
      
      /* first, wait for any outstanding operation to complete */
      wait_for_turn(c);
      
      /* check timestamp again */
      if (c->timestamp > USE_MACHINE(local_timestamp)[Process])
	goto allowed;
      
      CILK_ASSERT(!c->op_in_progress);
      c->op_in_progress = FETCH_IN_PROGRESS;
      
      Cilk_mutex_signal(&USE_MACHINE(cache_lock));
      
      pn = machine_hash(page, stack_id);
      req = Cilk_alloc_msg_buf(sizeof(FetchRequest), pn);
      req->page = page;
      req->stack_id = stack_id;
      req->machine = Machine;
      req->timestamp = USE_MACHINE(latest_timestamp);
      req->c = c;
      
      TRACE(TRACE_FETCH, ("requesting page %p(%d) from %d\n", page, stack_id, pn));
      
      Cilk_rpc(pn, Cilk_bstore_fetch_req, req, sizeof(*req));
      
      Cilk_enter_state(STATE_WAITING_FOR_FETCH);
      
      while(c->op_in_progress == FETCH_IN_PROGRESS)
	Cilk_poll(); /* yield()? KHR */
      
      Cilk_exit_state(STATE_WAITING_FOR_FETCH);
      
      Cilk_mutex_wait(&USE_MACHINE(cache_lock));
      
      /*
       * Now we have a page with a later timestamp - try
       * fetch loop again.  We need to try again because the
       * local_timestamp for this process may have gone up in
       * the meantime.
       */
      try++;
      goto again;
    }
  
allowed:
  if (try == 0)
    Cilk_event(EVENT_FETCH_NULL);
  
  set_bit(&c->read, Process);
  /* update_map(c); ???!!! */
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
  Cilk_map_cache_page(page, c->working, CILK_MAP_READONLY);
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  TRACE(TRACE_CACHE, ("mapped page %p(%d) readable\n", page, c->stack_id));
}

/*
 *  This code ignores the <write> flag for now.  If <write> is 1,
 * we should eventually map the page as writeable directly, so as
 * to avoid a second fault to map the page read/write.
 */
int Cilk_fault_handler(void *addr, int write)
{
  void *page;
  CachePage *c;
  void *work;
  int stack_id;
  
  if ((addr < USE_MACHINE(heap) || addr >= USE_MACHINE(heap) + USE_MACHINE(heap_size)) &&
      (addr < USE_PRIVATE(shared_region_begin) || addr >= USE_PRIVATE(shared_region_end)))
    return 0;
  
  page = addr - ((size_t)addr & (USE_MACHINE(pagesize)-1));
  stack_id = find_stack_id(page);
  
  TRACE(TRACE_CACHE, ("faulted on page %p(%d)\n", page, stack_id));
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  /*
   * check to see if page is in the cache already.
   */
  c = find_page(page, stack_id);
  if (c == NULL)
    {
      /*
       * Page is not in the cache yet.  Allocate a page first.
       */
      
      /* get working page */
      
      /* correct coloring policy - color of working page and color
       * of virtual page should be the same
       */
      work = get_cache_page(page_color(page));
      
      /*
       * Make sure page still isn't in cache - if it is,
       * free the new page we just got and break out of this case.
       */
      c = find_page(page, stack_id);
      if (c != NULL)
	{
	  /* return just-allocated page to free list */
	  free_cache_page(work);
	  
	  /* page may be mapped already, so we just retry -
	   * if page is not mapped, fault will happen again.
	   */
	  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
	  return 1;
	}
      
      /* get free header */
      c = USE_MACHINE(header_free_list);
      CILK_ASSERT(c != NULL);
      USE_MACHINE(header_free_list) = c->next;
      
      /* initialize cache page data */
      c->page = page;
      c->stack_id = stack_id;
      c->twin = NULL;
      c->working = work;
      c->read = EMPTY_BIT_VECTOR;
      c->write = EMPTY_BIT_VECTOR;
      c->dirty = EMPTY_BIT_VECTOR;
      c->op_in_progress = 0;
      c->timestamp = 1;
      c->pins = 0;
      //Cilk_mutex_init(&c->page_lock);
      
      /* splice cache entry into hash table */
      splice_in_hash(c, &USE_MACHINE(hasht)[cache_hash(page,stack_id)]);
      /* splice_in_lru(c, USE_MACHINE(unmapped_pages)); */
      splice_in_lru(c, USE_MACHINE(mapped_pages));
    }
  
  /* make sure page doesn't get moved or freed while we are working on it */
  c->pins++;
  
  CILK_ASSERT(!get_bit(&c->write, Process));
  
  if (get_bit(&c->read, Process))
    write_fault(c);
  else
    {
      read_fault(c);
#if 0
      if (write)
	{
	  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
	  /* make sure c is still in the cache! -KHR */
	  write_fault(c);
	}
#endif
    }
  
  c->pins--;
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
  
  TRACE(TRACE_CACHE, ("returning successfully from fault handler\n"));
  return 1;
}

typedef struct
{
  void *page;
  int stack_id;
} PageID;

/*
 * Commit all heap data to backing store.  Only guaranteed to reconcile
 * dirty pages that are in the cache at the start of this call.  Any
 * pages made dirty during this call may not be reconciled.  Only pages
 * dirty with respect to <proc> are reconciled.
 */
static void reconcile_cache(int proc)
{
  CachePage *c;
  int npages;
  PageID *pages;
  int i;
  
  TRACE(TRACE_CACHE, ("starting reconcile\n"));
  
  npages = USE_MACHINE(cache_pages);
  pages = alloca(npages * sizeof(PageID));
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  /*
   * Find out what pages we have to reconcile
   */
  i = 0;
  for(c=USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages); c=c->forw)
    if (get_bit(&c->dirty, proc))
      {
	pages[i].page = c->page;
	pages[i].stack_id = c->stack_id;
	i++;
      }
  CILK_ASSERT(i <= npages);
  npages = i;
  
  /*
   * Now, reconcile them
   */
  for(i=0; i<npages; i++)
    {
      c = find_page(pages[i].page, pages[i].stack_id);
      if (c == NULL)
	continue;                       /* already flushed */
      
      c->pins++;
      
      /*
       * wait for any previous fetch/reconcile to complete
       */
    again:
      wait_for_turn(c);
      
      if (!get_bit(&c->dirty, proc))
	{
	  c->pins--;
	  continue;    /* already reconciled */
	}
      
      /*
       * Map page read-only before reconciling, if we can.
       */
      if (proc == Process && get_bit(&c->write, proc))
	{
	  clear_bit(&c->write, proc);
	  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
	  Cilk_map_cache_page(c->page, c->working, CILK_MAP_READONLY);
	  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
	  goto again;
	}
      
      TRACE(TRACE_CACHE, ("reconcile page %p\n", c->page));
      
      reconcile_page(c);
      
      c->pins--;
    }
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
  
  TRACE(TRACE_CACHE, ("done reconcile\n"));
}

void Cilk_reconcile_cache(void)
{
  reconcile_cache(Process);
}
void Cilk_reconcile_cache_from(int proc)
{
  reconcile_cache(proc);
}

/* unmap any pages which are no longer accessable due to 
 * timestamp constraints.
 */
static void unmap_pages(void)
{
  CachePage *c;
  int npages;
  PageID *pages;
  int i;
  
  npages = USE_MACHINE(cache_pages);
  pages = alloca(npages * sizeof(PageID));
  
  /*
   * Record pages we can no longer access because of timestamp constraints.
   */
  i = 0;
  for(c=USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages); c=c->forw)
    {
      if (get_bit(&c->read, Process) &&
	  ((c->timestamp < USE_MACHINE(local_timestamp)[Process])))
	{
	  pages[i].page = c->page;
	  pages[i].stack_id = c->stack_id;
	  i++;
	}
    }
  CILK_ASSERT(i <= npages);
  npages = i;
  
  /*
   * Now, unmap them
   */
  for(i=0; i<npages; i++)
    {
      c = find_page(pages[i].page, pages[i].stack_id);
      if (c == NULL)
	continue;                       /* already flushed */
      
      if (get_bit(&c->read, Process) &&
	  (c->timestamp < USE_MACHINE(local_timestamp)[Process]))
	{
	  c->pins++;
	  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
	  Cilk_map_cache_page(c->page, NULL, CILK_MAP_FAULTING);
	  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
	  clear_bit(&c->read, Process);
	  clear_bit(&c->write, Process);
	  /* update_unmap(c); */
	  c->pins--;
	}
    }
}

/*
 * Do an acquire from one process to another within a machine
 */
void Cilk_cache_acquire_local_from(int proc)
{
  CachePage *c;
  
  TRACE(TRACE_CACHE, ("acqire local from %d\n", proc));
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  /*
   * Update timestamp of local process to be max of two
   * synchronizing processes
   */
  if (USE_MACHINE(local_timestamp)[Process] < USE_MACHINE(local_timestamp)[proc])
    USE_MACHINE(local_timestamp)[Process] = USE_MACHINE(local_timestamp)[proc];
  
  /*
   * Copy over dirty bits of <proc>
   */
  for(c=USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages); c=c->forw)
    {
      /* copy over dirty bits */
      if (get_bit(&c->dirty, proc))
	{
	  set_bit(&c->dirty, Process);
	  CILK_ASSERT(c->twin != NULL);
	}
    }
  
  unmap_pages();
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
}

/*
 * Same as above, but acquire from all processes on the machine.
 */
void Cilk_cache_acquire_local(void)
{
  CachePage *c;
  
  TRACE(TRACE_CACHE, ("acqire local\n"));
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  /*
   * Update timestamp of local process to be max of all processes in machine
   */
  USE_MACHINE(local_timestamp)[Process] = USE_MACHINE(latest_timestamp) - 1;
  
  /*
   * Copy over dirty bits from all processes.
   */
  for(c=USE_MACHINE(mapped_pages)->forw; c != USE_MACHINE(mapped_pages); c=c->forw)
    {
      /* copy over dirty bits */
      if (any_bits(&c->dirty))
	{
	  set_bit(&c->dirty, Process);
	  CILK_ASSERT(c->twin != NULL);
	}
    }
  
  unmap_pages();
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
}

/*
 * Unmap all pages.  Only guaranteed to unmap pages which are in
 * the cache at the start of this call.  Pages mapped in during
 * the call may remain.
 */
void Cilk_flush_cache(void)
{
  TRACE(TRACE_CACHE, ("flush cache\n"));
  
  Cilk_mutex_wait(&USE_MACHINE(cache_lock));
  
  /*
   * Allocate a new, largest timestamp so no current pages will be readable.
   */
  USE_MACHINE(local_timestamp)[Process] = USE_MACHINE(latest_timestamp) + 1;
  USE_MACHINE(latest_timestamp) += 2;
  
  unmap_pages();
  
  Cilk_mutex_signal(&USE_MACHINE(cache_lock));
}
