/*
 * $Id: cilk.c,v 1.3.2.8 1997/11/25 06:29:00 randall Exp $
 */

/*
 * $Log: cilk.c,v $
 * Revision 1.3.2.8  1997/11/25 06:29:00  randall
 * Reworked buffer management for memory channel.
 *
 * Revision 1.3.2.7  1997/11/12 23:13:48  randall
 * Memory channel is now working...
 *
 * Revision 1.3.2.6  1997/11/12 02:32:31  randall
 * More work on message-passing layer.
 *
 * Revision 1.3.2.5  1997/11/10 22:27:57  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.3.2.4  1997/10/21 23:02:40  randall
 * Separated out UDP implementation.  Reorganized caching algorithm.
 *
 * Revision 1.3.2.3  1997/10/14 20:25:58  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.3.2.2  1997/06/19 16:23:18  randall
 * Removed extraneous include.
 *
 * Revision 1.3.2.1  1997/06/13 22:33:17  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.7  1997/09/04 19:12:29  athena
 * Fixed error messages to be more personal and polite to the user.
 *
 * Revision 1.6  1997/09/04 17:38:05  bsong
 * Fixed the problem that the strings are expended with some macros, when
 * CILK_COMPLAIN happens.
 *
 * Revision 1.5  1997/09/03 23:00:54  bsong
 * Added some helpful error message for the user by using CILK_COMPLAIN.
 *
 * Revision 1.4  1997/08/25 19:58:54  athena
 * Added the consistency check for flags.
 * Stats level is now 0 by default: nothing is printed.
 *
 * Revision 1.3  1997/06/09 19:03:09  athena
 * Changed RTS so that Cilk won't dump core by default
 *
 * Revision 1.2  1997/03/01 17:42:44  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <stdio.h>

/* Hack: main is defined to be cilk_main; undef it */
#undef main

/* globally-visible options (read-only in child processes) */
DECLARE_PRIVATE_VARIABLE(int, active_size, );
DECLARE_PRIVATE_VARIABLE(int, yield_slice, );
DECLARE_MACHINE_VARIABLE(size_t, stackdepth, );

/*
 * variables that children need to update (typically at the end
 * of the program
 */
WHEN_CILK_TIMING(DECLARE_MACHINE_VARIABLE(Cilk_time, critical_path,);)
WHEN_CILK_TIMING(DECLARE_MACHINE_VARIABLE(Cilk_time, total_work, );)


/*
 * this string is printed when an assertion fails.
 * If we just inline it, apparently gcc generates many copies
 * of the string.
 */
DECLARE_MACHINE_VARIABLE(const char *, assertion_failed_msg, );
DECLARE_MACHINE_VARIABLE(const char *, stack_overflow_msg, );

/*
 * Random junk
 */
DECLARE_PRIVATE_VARIABLE(Cilk_time, wall_start, );

static char *smart_sprint_time(double x)
{
     static char buf[128];

     if (x < 1.0E-3)
	  sprintf(buf, "%f us", x * 1.0E6);
     else if (x < 1.0)
	  sprintf(buf, "%f ms", x * 1.0E3);
     else
	  sprintf(buf, "%f s", x);

     return buf;
}
/*
 * The statistics printer 
 */
static void print_all_statistics(Cilk_time start_time, Cilk_time end_time, const Options *options)
{
     int StatLevel = options->statlevel;
     int active_size;
     FILE *infofile;
     int i;

     active_size = 0;
     for(i=0; i<options->nmach; i++)
       active_size += Cilk_option_nproc[i];

     infofile = stderr;

     CILK_CHECK(infofile != NULL,
		("can't open stderr for writing\n"));

     if (StatLevel >= 1) {
	  /* level 1 and above: print wall clock, work and CP */
	  fprintf(infofile,
		  "Wall-clock running time on %d processor%s: %s\n",
		  active_size, 
		  active_size > 1 ? "s" : "",
		  smart_sprint_time(Cilk_wall_time_to_sec(
		       end_time - start_time)));
#if 0
	  WHEN_CILK_TIMING({
	       fprintf(infofile, "Total work = %s\n",
		       smart_sprint_time(Cilk_compute_work(options->nproc)));
	  });
#endif

	  /* 
	   * hack: we don't consider the total work if it
	   * is too small.  The trick allows us to compile
	   * the rts with CILK_CRITICAL_PATH=something and the
	   * user program with CILK_CRITICAL_PATH=0
	   */
#if 0
	  WHEN_CILK_TIMING({
	       if (Cilk_time_to_sec(USE_GLOBAL(total_work)) > 0.001) {
		    fprintf(infofile,
			    "Total work (accumulated) = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_GLOBAL(total_work))));
		    fprintf(infofile,
			    "Critical path = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_GLOBAL(critical_path))));
		    fprintf(infofile,
			    "Average parallelism = %f\n",
			    (double) USE_GLOBAL(total_work) /
			    (double) USE_GLOBAL(critical_path));
	       }
	  });
#endif
     }
     fclose(infofile);
}

#if CILK_MANAGE_PRIVATE_SEGMENT
/*
 * this is really useless, since if Cilk_private_vars
 * is really private, then all the other variables can
 * be private as well.  But at least the mechanism is
 * here, so that if Cilk_private_vars can be put in
 * some register or special location, everything works.
 */
CILK_DEFINE_REALLY_PRIVATE(struct Cilk_private_variables *, Cilk_private_vars);

static struct Cilk_private_variables dummy_private;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_map_private_segment(void)
{
     /* 
      * we use a dummy value because malloc() collects some statistics
      * that use private variables --- aaargh
      */
     Cilk_private_vars = &dummy_private;
     Self = -1;
     Cilk_private_vars = 
	  Cilk_malloc_fixed(sizeof (struct Cilk_private_variables));
     CILK_CHECK(Cilk_private_vars,
		("Can't allocate shared variables!\n"));
}

#endif

static void init_machine_vars(int id, const Options *options)
{
     USE_MACHINE(Machine) = id;
     USE_MACHINE(Processes) = options->permach;
     USE_MACHINE(Machines) = options->nmach;
     USE_MACHINE(stackdepth) = options->stackdepth;
     USE_MACHINE(done) = 0;
     USE_MACHINE(assertion_failed_msg) = 
                    "Assertion failed: %s line %d file %s\n"
		    "This is either a Cilk bug, or your program\n"
		    "has corrupted some Cilk internal data structure.\n";
     USE_MACHINE(stack_overflow_msg) =
		    "Cilk runtime system: Stack overflow.\n"
		    "Your program may have infinite recursion.\n";
}

static void init_private_vars(int id, const Options *options)
{
     USE_PRIVATE(Process) = id;
     Self = options->start_proc_no + id;

     USE_PRIVATE(active_size) = options->nproc;
     USE_PRIVATE(yield_slice) = options->yieldslice;
}


typedef struct
{
  int size;
  long data[0];
} PingPong;

#define DATA_CHECK 0
#define DATA_GEN 1
#define TWO_WAY 0

DECLARE_MACHINE_VARIABLE(int, ping_pong_flag, );

static void ping_reply(PingPong *reply)
{
#if DATA_CHECK && TWO_WAY
  {
    int i;
    for(i=0; i<reply->size/sizeof(long); i++)
      CILK_ASSERT(reply->data[i] == i);
  }
#endif
  USE_MACHINE(ping_pong_flag) = 0;
}

static void ping_req(PingPong *req)
{
  int size;
  PingPong *reply;
#if DATA_CHECK
  {
    int i;
    for(i=0; i<req->size/sizeof(long); i++)
      CILK_ASSERT(req->data[i] == i);
  }
#endif
#if TWO_WAY
  /* two way bandwidth */
  size = sizeof(PingPong) + req->size;
  reply = Cilk_alloc_msg_buf(size, 0);
#if DATA_GEN
  {
    int i;
    for(i=0; i<size/sizeof(long); i++)
      reply->data[i] = i;
  }
#endif
#else
  /* one way bandwidth */
  size = sizeof(PingPong);
  reply = Cilk_alloc_msg_buf(size, 0);
#endif
  Cilk_rpc(0, ping_reply, reply, size);
}

#define MIN(x,y) ((x)<(y) ? (x):(y))

static void ping_pong_test(void)
{
  int size,i;
  
  Cilk_dprintf("ping_pong test %d %d\n", USE_MACHINE(Machine), USE_PRIVATE(Process));
  
  CILK_ASSERT(USE_MACHINE(Machines) == 2);
  if (USE_MACHINE(Machine) != 0)
    while(1)
      {
	Cilk_poll();
      }
  if (USE_MACHINE(Machine) == 0)
    {
      for(size=8; size <= 2*1024*1024; size*=2)
	{
	  int n = MIN(100, 2*1024*1024/size);
	  
	  Cilk_time start = Cilk_get_time();
	  for(i=0; i<n; i++)
	    {
	      int msg_size = sizeof(PingPong) + size;
	      PingPong *req = Cilk_alloc_msg_buf(msg_size, 1);
	      req->size = size;
#if DATA_GEN
	      {
		int j;
		for(j=0; j<size/sizeof(long); j++)
		  req->data[j] = j;
	      }
#endif
	      USE_MACHINE(ping_pong_flag) = 1;
	      Cilk_rpc(1, ping_req, req, msg_size);
	      while(USE_MACHINE(ping_pong_flag))
		Cilk_poll();
	      /* Cilk_dprintf("%d\n", i); */
	    }
	  start = Cilk_get_time() - start;
	  Cilk_dprintf("n: %3d  size: %8d   round trip: %9.3f us   bw: %7.3f MB/s\n", n, size, Cilk_time_to_sec(start)*1e6/n, n * size / Cilk_time_to_sec(start) / 1e6);
	  start = Cilk_get_time();
	  while(Cilk_time_to_sec(Cilk_get_time() - start) < 1.0) Cilk_poll();
	}
    }
  else
    while(1)
      Cilk_poll();
}

/*
 * this function is the entry point for child processes
 */
static void child_main(const Options *options, int id)
{
     Closure *start;

     Cilk_map_private_segment();
     init_private_vars(id, options);
     
     Cilk_arch_specific_init_local(options);
     Cilk_memory_init_local(options);
     Cilk_malloc_init_local(options);
     Cilk_scheduler_init_local(options);
     Cilk_barrier_init_local(options);
     Cilk_message_passing_init_local(options);
     WHEN_CILK_STATS(Cilk_stats_init_local(options));
     WHEN_CILK_TIMING(Cilk_timing_init_local(options));

     Cilk_barrier();

     Cilk_init_timers();

     if (0)
       {
	 ping_pong_test();
	 return;
       }
     
     if (Self == 0)
          start = Cilk_create_initial_thread(options->argc, options->argv);
     else
          start = NULL;

     /*
      * Everything happens inside this call.
      */
     if (USE_PRIVATE(Process) == USE_MACHINE(Processes)-1) {
       while(!USE_MACHINE(done))
	 Cilk_poll();
     }
     else {
       Cilk_scheduler(start);
     }

     WHEN_CILK_STATS(Cilk_stats_done_local(options));
     WHEN_CILK_TIMING(Cilk_timing_done_local(options));
     Cilk_malloc_done_local(options);
     
     if (Self == 0)
       {
#if 0
	  WHEN_CILK_TIMING({
	       fprintf(stderr, "Total work = %s\n",
		       smart_sprint_time(
			    Cilk_time_to_sec(USE_MACHINE(total_work))));
	  });
	  WHEN_CILK_CRITICAL_PATH({
	       if (Cilk_time_to_sec(USE_MACHINE(total_work)) > 0.001) {
		    fprintf(stderr,
			    "Critical path = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_MACHINE(critical_path))));
		    fprintf(stderr,
			    "Average parallelism = %f\n",
			    (double) USE_MACHINE(total_work) /
			    (double) USE_MACHINE(critical_path));
	       }
	  });
#endif
       }
     
     /*
      * Make sure everything we printed gets to
      * the global process
      */
     fflush(stdout);
     fflush(stderr);
     
     Cilk_barrier();
     
     {
       char c = 4;
       write(1, &c, 1);
       write(2, &c, 1);
       while(1) Cilk_poll();        /* wait for kill from global process */
     }
}

void Cilk_machine_main(const Options *options, int id)
{
     Cilk_map_machine_segment();
     init_machine_vars(id, options);
     
     /* Cilk_malloc_machine_init(options); done in Cilk_memory_init_machine */
     Cilk_arch_specific_init_machine(options);
     Cilk_memory_init_machine(options);
     Cilk_malloc_init_machine(options);
     Cilk_scheduler_init_machine(options);
     Cilk_barrier_init_machine(options);
     Cilk_message_passing_init_machine(options);
     WHEN_CILK_STATS(Cilk_stats_init_machine(options));
     WHEN_CILK_TIMING(Cilk_timing_init_machine(options));
     
     /*
      * Everything happens inside this call.
      */
     Cilk_create_children(options, child_main);
     
     WHEN_CILK_STATS(Cilk_stats_done_machine(options));
     WHEN_CILK_TIMING(Cilk_timing_done_machine(options));
     Cilk_memory_done_machine(options);
     Cilk_malloc_machine_done(options);
     Cilk_arch_specific_done_machine(options);
}

/* global init */
static void global_main(Options *options)
{
     Cilk_arch_specific_init_global(options);
     Cilk_memory_init_global(options);
     WHEN_CILK_STATS(Cilk_stats_init_global(options));
     WHEN_CILK_TIMING(Cilk_timing_init_global(options));

     Cilk_message_passing_init_global(options);

#if 0
     /* initialize the statistics gatherer and timer */
     Cilk_time_gathering_init(options);
#endif
     
     /*
      * Everything happens inside this call.
      */
     Cilk_create_machines(options);

     WHEN_CILK_STATS(Cilk_stats_done_global(options));
     WHEN_CILK_TIMING(Cilk_timing_done_global(options));
     Cilk_memory_done_global(options);
}

/* magic consistency check */
int CILK_MAGIC_NAME;


int main(int argc, char *argv[])
{
     Cilk_time start_time, end_time;
     Options options;

     if (argc >= 1 && argv[1] != NULL && !strcmp(argv[1], "-__machine__"))
       {
	 /* we're already running, just starting local machine */
	 /* options will be read from master, not from command line */
	 /* remove -__machine__ option */
	 {
	   int i;
	   argc--;
	   for(i=1; i<argc; i++)
	     argv[i] = argv[i+1];
	 }
	 Cilk_start_machine(argc, argv);
	 exit(0);
       }

     start_time = Cilk_get_wall_time();

     /* parse command-line arguments */
     if (Cilk_parse_command_line(&options, argc, argv))
	  return 0;

     /* initialize and run! */
     global_main(&options);

     end_time = Cilk_get_wall_time();

     /* everybody is done: print stats and exit */
     print_all_statistics(start_time, end_time, &options);
     return 0;
}

