/*
 * $Id: sched.c,v 1.2.2.10 1997/12/22 20:52:21 randall Exp $
 */

/*
 * $Log: sched.c,v $
 * Revision 1.2.2.10  1997/12/22 20:52:21  randall
 * Fixed exit, alloca
 *
 * Revision 1.2.2.9  1997/12/18 23:38:43  randall
 * Added dummy abort code.
 *
 * Revision 1.2.2.8  1997/11/25 06:29:02  randall
 * Reworked buffer management for memory channel.
 *
 * Revision 1.2.2.7  1997/11/17 19:11:00  randall
 * Fixed up some alignment stuff.
 *
 * Revision 1.2.2.6  1997/11/12 23:13:50  randall
 * Memory channel is now working...
 *
 * Revision 1.2.2.5  1997/11/10 22:28:00  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.2.2.4  1997/11/02 00:15:59  randall
 * Merged in latest Cilk-5 stuff.
 *
 * Revision 1.2.2.3  1997/10/21 23:02:43  randall
 * Separated out UDP implementation.  Reorganized caching algorithm.
 *
 * Revision 1.2.2.2  1997/10/14 20:26:08  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.2.2.1  1997/06/13 22:33:24  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.7  1997/10/01 18:50:46  randall
 * Final fix.
 *
 * Revision 1.6  1997/10/01 18:48:33  randall
 * Fixed 64-bit random number generator bug.
 *
 * Revision 1.5  1997/09/17 19:07:34  randall
 * Added digital unix thread version.
 *
 * Revision 1.4  1997/09/04 19:12:35  athena
 * Fixed error messages to be more personal and polite to the user.
 *
 * Revision 1.3  1997/09/03 23:00:57  bsong
 * Added some helpful error message for the user by using CILK_COMPLAIN.
 *
 * Revision 1.2  1997/03/01 17:42:56  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * This file contains the scheduler and associated functions
 */

/* !!! remove this eventually ??? */
#include <stdio.h>

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <string.h>

DECLARE_MACHINE_VARIABLE(int, retval,);	/* return value from main */
DECLARE_MACHINE_VARIABLE(int, done,);

/* 
 * The ready deque.  Each ready queue contains a list of ready closures;
 * the last element of the deque may also be RUNNING or RETURNING
 */
typedef struct ReadyDeque {
     Cilk_mutex mutex;
     WHEN_CILK_DEBUG(int mutex_owner;)
     Closure *top, *bottom;
     CILK_CACHE_LINE_PAD;
} ReadyDeque;


DECLARE_PRIVATE_VARIABLE(ClosureCache *, cache ,);
DECLARE_PRIVATE_VARIABLE(int, _Self, );
DECLARE_PRIVATE_VARIABLE(int, Process, );
DECLARE_MACHINE_VARIABLE(int, Machine, );
DECLARE_MACHINE_VARIABLE(int, Processes, );
DECLARE_MACHINE_VARIABLE(int, Machines, );

WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, last_cp_time,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, cp_hack,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, work_hack,);)

DECLARE_PRIVATE_VARIABLE(Cilk_time, user_work,);
DECLARE_PRIVATE_VARIABLE(Cilk_time, user_critical_path,);

#define CLOSURE_HEAD(cl)  (cl->cache->head)
#define CLOSURE_EXCEPTION(cl)  (cl->cache->exception)
#define CLOSURE_TAIL(cl)  (cl->cache->tail)
#define CLOSURE_STACK(cl) (cl->cache->stack)
#define CLOSURE_ABORT(cl) (cl->abort_status)

#define NOBODY (-1)   /* invalid processor number */

/* 
 * SCHEDULER LOCK DIAGRAM
 * 
 * locks must be grabbed in the following order:
 *
 * ReadyDeque lock
 * Closure lock
 *
 * Closure locks must be acquired downwards
 * (parent first, then children)
 *
 * See also PROTOCOLS for a more verbose description.
 */

/*************************************************************
 * Basic operations on closures
 *************************************************************/
static inline void Closure_checkmagic(Closure *t)
{
     CILK_ASSERT(t->magic == CILK_CLOSURE_MAGIC);
}

/* assert that t be locked by ourselves */
static inline void Closure_assert_ownership(Closure *t)
{
     CILK_ASSERT(t->mutex_owner == Self);
}

/* assert that t be not locked */
static inline void Closure_assert_alienation(Closure *t)
{
     CILK_ASSERT(t->mutex_owner != Self);
}

static inline void Closure_destroy(Closure *t)
{
     Closure_checkmagic(t);

     WHEN_CILK_DEBUG(t->status = CLOSURE_DESTROYED);
     WHEN_CILK_DEBUG(t->magic = ~CILK_CLOSURE_MAGIC);
     Cilk_free_machine(t, sizeof(Closure));
}

static inline void Closure_lock(Closure *t)
{
     Closure_checkmagic(t);
     Cilk_mutex_wait(&(t->mutex));
     WHEN_CILK_DEBUG(t->mutex_owner = Self);
}

static inline void Closure_unlock(Closure *t)
{
     Closure_checkmagic(t);
     Closure_assert_ownership(t);
     WHEN_CILK_DEBUG(t->mutex_owner = -1);
     Cilk_mutex_signal(&(t->mutex));
}

static inline ClosureID Closure_make_ID(Closure *t)
{
  return (ClosureID){t, USE_MACHINE(Machine)};
}

/*************************************************************
 * Complicated operations on closures
 *************************************************************/

typedef struct
{
  ClosureID stealer;
  int child_num;
  Closure *parent;
  ReturnStuff stuff;   /* must be last - variable size */
} ReturnRequest;

typedef struct
{
  Closure *local_closure;
  Closure closure;
  int nretvals;
  int ninlets;
  char data[0];      /* extra data - returned values and incomplete inlets */
} StealReply;

static void pack_up_and_send_handler(StealReply *reply)
{
  Closure *t;
  int n;
  struct ReturnedValue *r;
  struct InletClosure *i;
  char *buf;
  
  t = reply->local_closure;
  
  Closure_lock(t);
  
  CILK_ASSERT(t->status == CLOSURE_NOT_VALID_YET);
  
  /* save setting the status to CLOSURE_READY until the rest of the
   * closure has been initialized.
   */
  CILK_ASSERT(reply->closure.status == CLOSURE_READY);
  reply->closure.status = CLOSURE_NOT_VALID_YET;
  
  /*
   * Hack to lock transferred closure in preparation of copyover
   */
  Cilk_mutex_init(&reply->closure.mutex);
  Closure_lock(&reply->closure);
  
  /*
   * Add in any return values closure has received already
   */
  CILK_ASSERT(reply->closure.returned_values == NULL);
  reply->closure.returned_values = t->returned_values;
  reply->closure.join_counter += t->join_counter;
  
  /*
   * overwrite closure with correct closure info
   */
  *t = reply->closure;
  
  buf = reply->data;
  
  /*
   * Unpack any transferred returned values
   */
  for(n=0; n<reply->nretvals; n++)
    {
      r = Cilk_malloc_machine(sizeof(struct ReturnedValue));
      memcpy(r, buf, sizeof(struct ReturnedValue));
      buf += sizeof(struct ReturnedValue);
      if (r->size)
	{
	  r->value = Cilk_malloc_machine(ALIGN(r->size, CILK_MAX_ALIGN));
	  memcpy(r->value, buf, r->size);
	  buf += ALIGN(r->size, CILK_MAX_ALIGN);
	}
      else
	{
	  WHEN_CILK_DEBUG(r->value = NULL);
	}
      r->next = t->returned_values;
      t->returned_values = r;
    }
  
  /*
   * Unpack any incomplete inlets
   */
  CILK_ASSERT(t->incomplete_inlets == NULL);
  for(n=0; n<reply->ninlets; n++)
    {
      i = Cilk_malloc_machine(sizeof(struct InletClosure));
      memcpy(i, buf, sizeof(struct InletClosure));
      buf += sizeof(struct InletClosure);
      if (i->argsize)
	{
	  i->inlet_args = Cilk_malloc_machine(ALIGN(i->argsize, CILK_MAX_ALIGN));
	  memcpy(i->inlet_args, buf, i->argsize);
	  buf += ALIGN(i->argsize, CILK_MAX_ALIGN);
	}
      else
	{
	  WHEN_CILK_DEBUG(i->inlet_args = NULL);
	}
      i->next = t->incomplete_inlets;
      t->incomplete_inlets = i;
    }
  
  /* now, set the status to CLOSURE_READY */
  Cilk_membar_StoreStore();
  t->status = CLOSURE_READY;
  
  Closure_unlock(t);
}

/*
 *  <t> is a closure to send to a remote location.  <t> should be locked.
 * <dest> is the ID of a closure that is currently in a CLOSURE_NOT_VALID_YET
 * state.  This routine transfers <t> to <dest>, leaving <t> as a
 * CLOSURE_FORWARDING closure that points to <dest>.
 */
static void pack_up_and_send(Closure *t, ClosureID dest, int home_proc)
{
  char *buf;
  struct ReturnedValue *r;
  struct InletClosure *i;
  StealReply *reply;
  int size;
  int nretvals, ninlets;
  
  CILK_ASSERT(dest.machine >= 0);
  Closure_assert_ownership(t);
  
  /*
   * pack up closure cache.  Shouldn't need to be done.
   */
  CILK_ASSERT(t->cache == NULL);
  CILK_ASSERT(t->next_ready == NULL);
  CILK_ASSERT(t->prev_ready == NULL);
  
  /*
   * make message to forward closure
   */
  size = sizeof(StealReply);
  nretvals = 0;
  for (r = t->returned_values; r; r = r->next)
    {
      nretvals++;
      size += sizeof(struct ReturnedValue) + ALIGN(r->size, CILK_MAX_ALIGN);
    }
  ninlets = 0;
  for (i = t->incomplete_inlets; i; i = i->next)
    {
      ninlets++;
      size += sizeof(struct InletClosure) + ALIGN(i->argsize, CILK_MAX_ALIGN);
    }
  
  reply = Cilk_alloc_msg_buf(size, dest.machine);
  reply->local_closure = dest.c;
  reply->nretvals = nretvals;
  reply->ninlets = ninlets;
  
  buf = reply->data;
  
  /*
   * pack up return values
   */
  for(r = t->returned_values; r; r = t->returned_values)
    {
      t->returned_values = r->next;
      
      memcpy(buf, r, sizeof(struct ReturnedValue));
      buf += sizeof(struct ReturnedValue);
      if (r->size)
	{
	  memcpy(buf, r->value, ALIGN(r->size, CILK_MAX_ALIGN));
	  buf += ALIGN(r->size, CILK_MAX_ALIGN);
	  Cilk_free_machine(r->value, ALIGN(r->size, CILK_MAX_ALIGN));
	}
      Cilk_free_machine(r, sizeof(struct ReturnedValue));
    }

  /*
   * pack up incomplete inlets
   */
  for (i = t->incomplete_inlets; i; i = t->incomplete_inlets)
    {
      t->incomplete_inlets = i->next;
      
      memcpy(buf, i, sizeof(struct InletClosure));
      buf += sizeof(struct InletClosure);
      if (i->argsize)
	{
	  memcpy(buf, i->inlet_args, ALIGN(i->argsize, CILK_MAX_ALIGN));
	  buf += ALIGN(i->argsize, CILK_MAX_ALIGN);
	  Cilk_free_machine(i->inlet_args, ALIGN(i->argsize, CILK_MAX_ALIGN));
	}
      Cilk_free_machine(i, sizeof(struct InletClosure));
    }
  
  /*
   * pack up closure
   */
  reply->closure = *t;
  /* Unlock lock in message buffer - cheating, but it works */
  WHEN_CILK_DEBUG(reply->closure.mutex_owner = -1);
  Cilk_mutex_signal(&reply->closure.mutex);
  
  /*
   * Make old closure into forwarder
   */
  if (t->join_counter)
    {
      t->status = CLOSURE_FORWARDING;
      t->forward = dest;
      CILK_ASSERT(dest.machine >= 0);
      Closure_unlock(t);
    }
  else
    Closure_destroy(t);
  
  /*
   * Make sure place closure is going to gets updated memory
   */
  Cilk_memory_release_from(NULL, 0, home_proc, 2);
  
  /*
   * Send message to machine requesting closure
   */
  /* Cilk_dprintf("sending closure\n"); */
  Cilk_rpc(dest.machine, pack_up_and_send_handler, reply, size);
}


/*************************************************************
 * Sending steal failures
 *************************************************************/

typedef struct
{
  Closure *closure;
} StealFailure;

static void steal_failure_handler(StealFailure *reply)
{
  Closure *t = reply->closure;
  Cilk_fence();
  
  CILK_ASSERT(t->status == CLOSURE_NOT_VALID_YET);
  t->status = CLOSURE_STEAL_FAILURE;
}

/*
 * Send a steal failure to someone waiting for a steal.  The waiter
 * must have set up <closure> to be in the CLOSURE_NOT_VALID_YET
 * state.
 */
static void send_steal_failure(ClosureID closure)
{
  StealFailure *reply = Cilk_alloc_msg_buf(sizeof(StealFailure), closure.machine);
  reply->closure = closure.c;
  Cilk_rpc(closure.machine, steal_failure_handler, reply, sizeof(*reply));
}

/***********************************************************
 *  Internal random number generator, since rand()/random() is not
 *  MT-Safe. This RNG is *distinct* from the user RNG, just 
 *  for safety.
 ***********************************************************/
/* code stolen from NetBSD */

/*
 * Cilk_rand_next is stored in the per-processor global section to
 * make this code MT-Safe.
 */
DECLARE_PRIVATE_VARIABLE(static unsigned int, rand_next, );

static unsigned int rts_rand(void)
{
     USE_PRIVATE(rand_next) = USE_PRIVATE(rand_next) * 1103515245 + 12345;
     return (USE_PRIVATE(rand_next) >> 16);
}

static void rts_srand(unsigned int seed)
{
     USE_PRIVATE(rand_next) = seed;
}

/*********************************************************
 * Management of ReadDeques
 *********************************************************/

/* dynamically-allocated array of deques, one per processor */
DECLARE_MACHINE_VARIABLE(static struct ReadyDeque *, deques, );

/* initialize the deques */
static void init_deques(const Options *options)
{
     int i;

     USE_MACHINE(deques) = Cilk_malloc_machine_fixed(options->permach *
						     sizeof(ReadyDeque));
     CILK_CHECK(USE_MACHINE(deques), ("failed to allocate deques\n"));

     for (i = 0; i < options->permach; ++i) {
	  USE_MACHINE(deques)[i].top = (Closure *) NULL;
	  USE_MACHINE(deques)[i].bottom = (Closure *) NULL;
	  Cilk_mutex_init(&USE_MACHINE(deques)[i].mutex);
	  WHEN_CILK_DEBUG(USE_MACHINE(deques)[i].mutex_owner = NOBODY);
     }
}

/* assert that pn's deque be locked by ourselves */
static inline void deque_assert_ownership(int pn)
{
     CILK_ASSERT(USE_MACHINE(deques)[pn].mutex_owner == Self);
}

static inline void deque_assert_alienation(int pn)
{
     CILK_ASSERT(USE_MACHINE(deques)[pn].mutex_owner != Self);
}

static inline void deque_lock(int pn)
{
     Cilk_mutex_wait(&USE_MACHINE(deques)[pn].mutex);
     WHEN_CILK_DEBUG(USE_MACHINE(deques)[pn].mutex_owner = Self);
}

static inline void deque_unlock(int pn)
{
     WHEN_CILK_DEBUG(USE_MACHINE(deques)[pn].mutex_owner = NOBODY);
     Cilk_mutex_signal(&USE_MACHINE(deques)[pn].mutex);
}

/* 
 * functions that add/remove elements from the top/bottom
 * of deques
 */
static Closure *deque_xtract_top(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_MACHINE(deques)[pn].top;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
	  USE_MACHINE(deques)[pn].top = cl->next_ready;
	  if (cl == USE_MACHINE(deques)[pn].bottom) {
	       CILK_ASSERT(cl->next_ready == (Closure *) NULL);
	       USE_MACHINE(deques)[pn].bottom = (Closure *) NULL;
	  } else {
	       CILK_ASSERT(cl->next_ready);
	       (cl->next_ready)->prev_ready = (Closure *) NULL;
	       WHEN_CILK_DEBUG(cl->next_ready = NULL);
	  }
	  WHEN_CILK_DEBUG(cl->owner_ready_deque = NOBODY);
     } else {
	  CILK_ASSERT(USE_MACHINE(deques)[pn].bottom == (Closure *) NULL);
     }
     
     return cl;
}

static Closure *deque_peek_top(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_MACHINE(deques)[pn].top;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
     } else {
	  CILK_ASSERT(USE_MACHINE(deques)[pn].bottom == (Closure *) NULL);
     }

     return cl;
}

static Closure *deque_xtract_bottom(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_MACHINE(deques)[pn].bottom;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
	  USE_MACHINE(deques)[pn].bottom = cl->prev_ready;
	  if (cl == USE_MACHINE(deques)[pn].top) {
	       CILK_ASSERT(cl->prev_ready == (Closure *) NULL);
	       USE_MACHINE(deques)[pn].top = (Closure *) NULL;
	  } else {
	       CILK_ASSERT(cl->prev_ready);
	       (cl->prev_ready)->next_ready = (Closure *) NULL;
	  }

	  WHEN_CILK_DEBUG(cl->owner_ready_deque = NOBODY);
     } else {
	  CILK_ASSERT(USE_MACHINE(deques)[pn].top == (Closure *) NULL);
     }
     
     return cl;
}

static Closure *deque_peek_bottom(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_MACHINE(deques)[pn].bottom;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
     } else {
	  CILK_ASSERT(USE_MACHINE(deques)[pn].top == (Closure *) NULL);
     }
     
     return cl;
}

#if 0   
/*
 * Unused code, that may be useful if you hack the scheduler.
 */
static void deque_add_top(Closure *cl, int pn)
{
     deque_assert_ownership(pn);
     CILK_ASSERT(cl->owner_ready_deque == NOBODY);

     cl->next_ready = USE_MACHINE(deques)[pn].top;
     cl->prev_ready = (Closure *)NULL;
     USE_MACHINE(deques)[pn].top = cl;
     WHEN_CILK_DEBUG(cl->owner_ready_deque = pn);
     
     if (USE_MACHINE(deques)[pn].bottom) {
	  CILK_ASSERT(cl->next_ready);
	  (cl->next_ready)->prev_ready = cl;
     } else {
	  USE_MACHINE(deques)[pn].bottom = cl;
     }
}
#endif

static void deque_add_bottom(Closure *cl, int pn)
{
     deque_assert_ownership(pn);
     CILK_ASSERT(cl->owner_ready_deque == NOBODY);

     cl->prev_ready = USE_MACHINE(deques)[pn].bottom;
     cl->next_ready = (Closure *)NULL;
     USE_MACHINE(deques)[pn].bottom = cl;
     WHEN_CILK_DEBUG(cl->owner_ready_deque = pn);
     
     if (USE_MACHINE(deques)[pn].top) {
	  CILK_ASSERT(cl->prev_ready);
	  (cl->prev_ready)->next_ready = cl;
     } else {
	  USE_MACHINE(deques)[pn].top = cl;
     }
}

static inline void deque_assert_is_bottom(Closure *t)
{
     deque_assert_ownership(Self);
     CILK_ASSERT(t == deque_peek_bottom(Self));
}


/*********************************************************
 * Management of ClosureCache's
 *********************************************************/

/* dynamically-allocated array of ClosureCache's, one per processor */
DECLARE_MACHINE_VARIABLE(static ClosureCache *, caches, );

/* initialize the caches */
static void init_caches(const Options *options)
{
     int i;

     /* CILK_ASSERT(Self == NOBODY); */
     USE_MACHINE(caches) = 
          Cilk_malloc_machine_fixed(options->permach * 
				    sizeof(ClosureCache));
     CILK_CHECK(USE_MACHINE(caches), ("failed to allocate caches\n"));

     for (i = 0; i < options->permach; ++i) {
	  USE_MACHINE(caches)[i].stack =
	       Cilk_malloc_machine_fixed(USE_MACHINE(stackdepth) *
					 sizeof(StackFrame *));
	  CILK_CHECK(USE_MACHINE(caches)[i].stack,
		     ("failed to allocate stack\n"));
     }
}

static inline ClosureCache *cache_get_addr(int pn)
{
     return USE_MACHINE(caches) + pn;
}

/**************************************************
 * Management of ProcInfo's
 **************************************************/
/*
 * Returns a pointer to the ProcInfoT structure describing entry <entry>
 * of procedure whose signature is p.
 */
static inline ProcInfo *get_ProcInfo(ProcInfo * p, int entry)
{
     CILK_ASSERT(entry > 0);
     return p + entry;
}

/*
 * Returns a pointer to the slow version for a procedure
 * whose signature is p.
 */
static void (*get_proc_slow(ProcInfo *p)) () {
     return p[0].inlet;
}

/********************************************
 * Closure management
 ********************************************/
static inline int Closure_at_top_of_stack(Closure *cl)
{
     return (CLOSURE_HEAD(cl) + 1 == CLOSURE_TAIL(cl));
}

static inline int Closure_has_children(Closure *cl)
{
     return (cl->join_counter != 0);
}

Closure *Cilk_Closure_create(void)
{
     Closure *new = Cilk_malloc_machine(sizeof(Closure));

     CILK_CHECK(new, ("can't allocate Closure\n"));

     Cilk_mutex_init(&new->mutex);

     new->cp = (Cilk_time) 0;
     new->work = (Cilk_time) 0;

     new->returned_values = (struct ReturnedValue *) NULL;
     new->incomplete_inlets = (struct InletClosure *) NULL;
     new->next_ready = (struct Closure *) NULL;
     new->prev_ready = (struct Closure *) NULL;
     CLOSURE_ABORT(new) = NO_ABORT;
     new->return_stuff = (void *)0;
     WHEN_CILK_DEBUG(new->owner_ready_deque = NOBODY);
     new->join_counter = 0;
     new->acquire_level_needed = 1;

     WHEN_CILK_DEBUG({
	  new->magic = CILK_CLOSURE_MAGIC;
	  new->owner_ready_deque = NOBODY;
	  new->mutex_owner = NOBODY;
     });

     Cilk_event(EVENT_CLOSURE_CREATE);

     return new;
}


/* 
 * if (entry >= 0)
 *      get the inlet according to the entry
 * else
 *      get the inlet from the frame
 */
static struct InletClosure *make_incomplete_inlet_closure(StackFrame *f)
{
     struct InletClosure *p;
     ProcInfo *info;

     p = (struct InletClosure *)
       Cilk_malloc_machine(sizeof(struct InletClosure));

     info = get_ProcInfo(f->sig, f->entry);

     /* if no index is statically specified, get it from the frame */
     if (info->index == -1)
	  p->index = f->index;
     else
          p->index = info->index;

     /* take care of inlets */
     if (info->inlet != NULL) {
	  /* get arguments of inlet out of frame and into closure */
          p->inlet = info->inlet;
          p->argsize = info->argsize;
	  p->inlet_args = Cilk_malloc_machine(ALIGN(info->argsize, CILK_MAX_ALIGN));

	  memcpy((void *) p->inlet_args,
		 (char *) f + info->argindex, info->argsize);
     } else {
	  /*
	   * non-inlet case
	   */
	  p->inlet = NULL;
	  p->argsize = 0;
	  p->inlet_args = NULL;
     }
     
     return p;
}

/* 
 * only the scheduler is allowed to alter the closure
 * tree.  Consequently, these operations are private
 */
static void Closure_add_child(Closure *parent, int child_num)
{
     struct InletClosure *p;

     Closure_assert_alienation(parent);

     p = make_incomplete_inlet_closure(parent->frame);

     /*
      * semantic overloading: the inlet closure also contains
      * the list of children.
      */
     p->child_num = child_num;

     Closure_lock(parent);
     Closure_checkmagic(parent);

     /* splice into list of unreturned children in parent */
     p->next = parent->incomplete_inlets;
     parent->incomplete_inlets = p;

     Closure_unlock(parent);
}

/*******************************************
 * Exceptions
 *******************************************/
/*
 * Cilk supports a clever way to send an interruption/exception
 * to a running worker, without interrupting its normal way
 * of working. The idea is as follows.
 *
 * The local worker operates at the bottom of the frame
 * stack, using a stack pointer T (tail).  The exception
 * mechanism maintains another pointer E (exception).
 * The worker polls the value of E when decrementing T.
 * Whenever T becomes less or equal than E, the worker
 * calls the exception handler of the runtime system,
 * that does all the necessary bookkeeping.
 *
 * Exceptions can be sent only to running closures.
 * The exception pointer is always greater or equal than
 * the head pointer (the index of the first non-stolen
 * frame on the stack)
 */

/* value of E bigger than any possible value of T */
#define EXCEPTION_INFINITY (volatile StackFrame**)0xFFFFFFFF

static inline void increment_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     if (CLOSURE_EXCEPTION(cl) != EXCEPTION_INFINITY) {
	  ++CLOSURE_EXCEPTION(cl);
	  
	  /* make sure the exception is visible, before we continue */
	  Cilk_fence();
     }
}

static inline void decrement_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     if (CLOSURE_EXCEPTION(cl) != EXCEPTION_INFINITY)
	  --CLOSURE_EXCEPTION(cl);
}

static inline void reset_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CLOSURE_EXCEPTION(cl) = CLOSURE_HEAD(cl);
}

static inline void signal_immediate_exception(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     CLOSURE_EXCEPTION(cl) = EXCEPTION_INFINITY;

     /* make sure the exception is visible, before we continue */
     Cilk_fence();
}

/***********************************************************
 *         Work-stealing and related functions
 ***********************************************************/
/* 
 * For stealing we use a Dekker-like protocol, that achieves
 * mutual exclusion using shared memory.
 *
 * Thief:
 *   lock
 *   E++          ; signal intention to steal, sending an exception
 *   (implicit MEMBAR)
 *   If (H >= T)  ; if we can steal
 *     E--        ; retract the exception
 *     give up
 *   steal (that is, H++, etc)
 *   unlock
 *
 * The victim does the normal exception-handling mechanism:
 *
 *   T--
 *   --MEMBAR-- ensure T write occurs before E is read.
 *   if (E>=T)
 *     
 *
 */

/*
 * Do the thief part of Dekker's protocol.  Return 1 upon success,
 * 0 otherwise.  The protocol fails when the victim already popped
 * T so that E=T.
 */
static int do_dekker_on(Closure *cl)
{
     Closure_assert_ownership(cl);

     increment_exception_pointer(cl);
     Cilk_membar_StoreLoad(); 

     if ((CLOSURE_HEAD(cl) + 1) >= CLOSURE_TAIL(cl)) {
	  decrement_exception_pointer(cl);
	  return 0;
     }

     return 1;
}

static void Closure_make_ready(Closure *cl)
{
     cl->cache = (ClosureCache *)NULL;
     cl->status = CLOSURE_READY;
}

/*
 * promote the child frame of parent to a full closure.
 *
 * Assumptions: the parent is running on victim, and we own
 * the locks of both parent and deque[victim].
 * The child keeps running on the same cache of the parent.
 *
 * In order to promote a child frame to a closure,
 * the parent's frame must be the last in its ready queue.
 */
static void promote_child_and_detach(Closure *parent, int victim)
{
     Closure *child = Cilk_Closure_create();
     Closure *tmp;

     Closure_assert_ownership(parent);
     deque_assert_ownership(victim);
     CILK_ASSERT(parent->status == CLOSURE_RUNNING);
     CILK_ASSERT(parent->owner_ready_deque == victim);
     CILK_ASSERT(parent->next_ready == (Closure *) NULL);

     /* can't promote until we are sure nobody works on the frame */
     CILK_ASSERT(CLOSURE_HEAD(parent) <= CLOSURE_EXCEPTION(parent));

     /* update the various fields */
     child->parent = Closure_make_ID(parent);
     child->child_num = parent->curr_child++;
     child->curr_child = 1;
     child->join_counter = 0;
     child->cache = parent->cache;
     child->status = CLOSURE_RUNNING;

     /* join counter update */
     parent->join_counter++;
     
     /* setup the frame of the child closure */
     CLOSURE_HEAD(child)++;
     child->frame = (StackFrame*)*CLOSURE_HEAD(child);

     /* insert the closure on the victim processor's deque */
     deque_add_bottom(child, victim);

     /* at this point the child can be freely executed */

     /* detach the parent */
     tmp = deque_xtract_top(victim);
     CILK_ASSERT(tmp == parent);
}

/*
 * Finishes the promotion process.  The child is already fully promoted
 * and requires no more work (we only use the given pointer to identify
 * the child).  This function does some more work on the parent to make
 * the promotion complete.
 */
void finish_promote(Closure *parent, int child_num)
{
     /* register this child */
     Closure_add_child(parent, child_num);

     return;
}

/*
 * stealing protocol.  Tries to steal from the victim; returns a
 * stolen closure, or NULL if none.  The closure is returned locked
 * if it is not NULL.
 */
static Closure *Closure_steal_local_thehardpart(int victim)
{
     Closure *res = (Closure *) NULL;
     Closure *cl;

     /* first attempt to see if any steal is possible.  If there
      * are one or zero closures in the deque, steal is not possible.
      */
     {
       ClosureCache *cache = cache_get_addr(victim);
       if (cache->head >= cache->tail - 1)
	 return res;
     }

     Cilk_event(EVENT_STEAL_ATTEMPT);

     deque_lock(victim);

     cl = deque_peek_top(victim);

     if (cl) {
	  Closure_lock(cl);
	  
	  switch (cl->status) {
	      case CLOSURE_READY:
		   Cilk_die("BUG: can't happen in Closure_steal\n");
		   break;

	      case CLOSURE_RUNNING:
		   /* send the exception to the worker */
		   if (do_dekker_on(cl)) {
			/* 
			 * if we could send the exception, promote
			 * the child to a full closure, and steal
			 * the parent
			 */
		        promote_child_and_detach(cl, victim);
			deque_unlock(victim);
			/* 
			 * at this point, more steals can happen
			 * from the victim.
			 */
			Closure_make_ready(cl);
			res = cl;
			Cilk_event(EVENT_STEAL);
		   } else {
			/* it is ok to swap these two */
		        Closure_unlock(cl);
			deque_unlock(victim);
			Cilk_event(EVENT_STEAL_NO_DEKKER);
		   }
		   break;
		   
	      case CLOSURE_SUSPENDED:
		   Cilk_die("Bug: suspended closure in ready deque\n");
		   break;

	      case CLOSURE_RETURNING:
		   /* ok, let it leave alone */
		   /* 
		    * MUST unlock the closure before the queue;
		    * see rule D in the file PROTOCOLS
		    */
		   Closure_unlock(cl);
		   deque_unlock(victim);
		   Cilk_event(EVENT_STEAL_RETURNING);
		   break;

	      default:
		   Cilk_die("Bug: unknown closure status\n");
	  }

     } else {
          deque_unlock(victim);
	  Cilk_event(EVENT_STEAL_EMPTY_DEQUE);
     }

     return res;
}

static Closure *Closure_steal_local(int *victim)
{
  Closure *res;

  *victim = rts_rand() % USE_MACHINE(Processes);
  res = Closure_steal_local_thehardpart(*victim);

  return res;
}

typedef struct
{
  ClosureID stealer;
} StealRequest;

static void steal_request_handler(StealRequest *req)
{
  Closure *t;
  int victim_process;
  
  t = Closure_steal_local(&victim_process);
  
  if (t == NULL)
    {
      /*
       * steal failure
       */
      /* Cilk_dprintf("sending steal failure to %d\n", req->stealer.machine); */
      send_steal_failure(req->stealer);
    }
  else
    {
      /*
       * steal success
       */
      
      /* Cilk_dprintf("sending steal success to %d\n", req->stealer.machine); */
      pack_up_and_send(t, req->stealer, victim_process);
    }
}

#define LOCAL_STEALS_PER_GLOBAL_STEAL 100
DECLARE_MACHINE_VARIABLE(static Cilk_mutex, remote_steal_lock, );

static Closure *Closure_steal(void)
{
  Closure *res;
  int victim_machine;
  
  if (rts_rand() % LOCAL_STEALS_PER_GLOBAL_STEAL == 0 &&
      Cilk_mutex_try(&USE_MACHINE(remote_steal_lock)))
    {
      victim_machine = rts_rand() % USE_MACHINE(Machines);
      if (victim_machine == USE_MACHINE(Machine))
	Cilk_mutex_signal(&USE_MACHINE(remote_steal_lock));
    }
  else
    victim_machine = USE_MACHINE(Machine);

  if (victim_machine == USE_MACHINE(Machine))
    {
      int victim_process;
      res = Closure_steal_local(&victim_process);
      if (res)
	{
	  Closure_unlock(res);
	  
	  Cilk_memory_acquire_from(NULL, 0, victim_process, 1);
	  
	  finish_promote(res, res->curr_child - 1);
	}
    }
  else
    {
      StealRequest *req = Cilk_alloc_msg_buf(sizeof(StealRequest), victim_machine);
      res = Cilk_Closure_create();
      res->status = CLOSURE_NOT_VALID_YET;
      
      req->stealer = Closure_make_ID(res);
      
      Cilk_rpc(victim_machine, steal_request_handler, req, sizeof(*req));
      
      while(res->status == CLOSURE_NOT_VALID_YET && !USE_MACHINE(done))
	Cilk_poll();
      
      Closure_lock(res);
      
      if (USE_MACHINE(done))
	{
	  /*
	   * We just count this closure as deallocated,
	   * even though we don't deallocate it, because a
	   * steal reply may still come in.
	   */
	  Cilk_pretend_to_free_machine(res, sizeof(Closure));
	  res = NULL;
	}
      else if (res->status == CLOSURE_STEAL_FAILURE)
	{
	  /* Cilk_dprintf("received steal failure\n"); */
	  Closure_destroy(res);
	  res = NULL;
	}
      else
	{
	  /* Cilk_dprintf("received steal success\n"); */
	  CILK_ASSERT(res->status == CLOSURE_READY);
	  /* CILK_ASSERT(res->magic == CILK_CLOSURE_MAGIC); */
	  
	  res->acquire_level_needed = 1;
	  Closure_unlock(res);
	  
	  Cilk_memory_acquire(NULL, 0, 2);
	  
	  finish_promote(res, res->curr_child - 1);
	}
      Cilk_mutex_signal(&USE_MACHINE(remote_steal_lock));
      if (res != NULL)
	Cilk_event(EVENT_STEAL_REMOTE_SUCCESS);
      else
	Cilk_event(EVENT_STEAL_REMOTE_FAILURE);
    }
  return res;
}

/********************************************************
 * Functions that take care of argument sending.
 ********************************************************/
static void apply_inlet(Closure *cl, struct InletClosure *i, struct ReturnedValue *r)
{
     int index = i->index;
     void (*inlet) (void *, void *, void *) = i->inlet;
     StackFrame *frame = cl->frame;

     Closure_assert_alienation(cl);

     if (inlet) {
	  /* run the inlet */
          int argsize = i->argsize;
          void *inlet_args = i->inlet_args;
	  
	  CILK_ASSERT(r->size <= argsize);
          if (r->size)
	    memcpy(inlet_args, r->value, r->size);
	  inlet(frame, inlet_args, (char *) frame + index);
	  if (argsize)
	    Cilk_free_machine(inlet_args, ALIGN(argsize, CILK_MAX_ALIGN));
     } else {
	  /* 
	   * default inlet: copy just-returned value into frame 
	   */
	  if (r->size)
	    {
	      CILK_ASSERT(index >= sizeof(StackFrame));
	      memcpy((char *) frame + index, r->value, r->size);
	    }
     }

     /* free the inlet structure and return value */
     Cilk_free_machine(i, sizeof(struct InletClosure));
     if (r->size)
       Cilk_free_machine(r->value, ALIGN(r->size, CILK_MAX_ALIGN));
     Cilk_free_machine(r, sizeof(struct ReturnedValue));
}

/* walk the list of inlets and execute all of them */
static void poll_inlets(Closure *cl)
{
     struct InletClosure *i, **ip;
     struct ReturnedValue *r;

     Closure_assert_alienation(cl);
     
     Cilk_event(EVENT_POLL_INLETS);

loop:
     Closure_lock(cl);
     r = cl->returned_values;
     if (r != NULL)
       {
	 cl->returned_values = r->next;
	 
	 for(ip = &cl->incomplete_inlets, i=*ip; i; ip = &i->next, i=*ip)
	   {
	     if (i->child_num == r->child_num)
	       {
		 *ip = i->next;
		 break;
	       }
	   }
	 CILK_ASSERT(i != NULL);
	 
	 /* update critical path and work */
	 WHEN_CILK_TIMING({
	   if (r->cp > cl->cp)
	     cl->cp = r->cp;
	   cl->work += r->work;
	 });
	 
	 Closure_unlock(cl);
	 
	 apply_inlet(cl, i, r);
	 
	 goto loop;
       }
     
     Closure_unlock(cl);
}

static void enque_return_value(Closure *parent, int child_num,
			       ReturnStuff *stuff)
{
     struct ReturnedValue *retval;
     Closure_assert_ownership(parent);
     Cilk_event(EVENT_RETURN_ENQUEUE);

     retval = Cilk_malloc_machine(sizeof(struct ReturnedValue));
     retval->child_num = child_num;
     WHEN_CILK_TIMING(retval->cp = stuff->cp);
     WHEN_CILK_TIMING(retval->work = stuff->work);
     retval->size = stuff->value_size;
     if (stuff->value_size > 0)
       {
	 retval->value = Cilk_malloc_machine(ALIGN(stuff->value_size, CILK_MAX_ALIGN));
	 memcpy(retval->value, stuff->value, ALIGN(stuff->value_size, CILK_MAX_ALIGN));
       }
     retval->next = parent->returned_values;
     parent->returned_values = retval;
}

static Closure *provably_good_steal_maybe(Closure *parent)
{
     Closure *res = (Closure *) NULL;

     Closure_assert_ownership(parent);

     if (parent->join_counter == 0 && parent->status == CLOSURE_SUSPENDED) {
	  /* do a provably-good steal; this is *really* simple */
	  res = parent;
	  Cilk_event(EVENT_PROVABLY_GOOD_STEAL);

	  /* debugging stuff */
	  CILK_ASSERT(parent->owner_ready_deque == NOBODY);
	  Closure_make_ready(parent);
     }

     return res;
}

/*
 * Requires that parent is locked.  Unlocks parent.  Returns
 * a provably-good steal, if any, in an unlocked state.
 */
static Closure *Closure_return_local(Closure *parent, int child_num,
				     ReturnStuff *stuff)
{
  Closure *res;

  Closure_assert_ownership(parent);

  CILK_ASSERT(parent->status != CLOSURE_RETURNING);
  /* CILK_ASSERT(parent->frame); */
  /* CILK_ASSERT(parent->frame->magic == CILK_STACKFRAME_MAGIC);  - can't access the frame here -KHR */
  
  enque_return_value(parent, child_num, stuff);
  
  CILK_ASSERT(parent->join_counter > 0 ||
	      (parent->join_counter <= 0 &&
	       parent->status == CLOSURE_NOT_VALID_YET));
  --parent->join_counter;
  
  res = provably_good_steal_maybe(parent);
  
  Closure_unlock(parent);
  
  if (res != NULL)
    Cilk_memory_acquire(NULL, 0, 1);
  
  return res;
}

static void return_request_handler(ReturnRequest *req)
{
  Closure *parent;
  
  parent = req->parent;
  
  Closure_lock(parent);
  
  if(parent->status == CLOSURE_FORWARDING)
    {
      /*
       * Forward request to where the closure went.
       */
      ReturnRequest *req2;
      ClosureID forward;
      int msg_size;
      
      forward = parent->forward;
      parent->join_counter--;
      if (parent->join_counter == 0)
	Closure_destroy(parent);
      else
	Closure_unlock(parent);
      
      msg_size = sizeof(ReturnRequest) + ALIGN(req->stuff.value_size, CILK_MAX_ALIGN);
      req2 = Cilk_alloc_msg_buf(msg_size, forward.machine);
      req2->stealer = req->stealer;
      req2->child_num = req->child_num;
      req2->parent = forward.c;
      req2->stuff = req->stuff;
      memcpy(req2->stuff.value, req->stuff.value, ALIGN(req->stuff.value_size, CILK_MAX_ALIGN));
      
      /* Cilk_dprintf("forwarding return request\n"); */
      
      Cilk_rpc(forward.machine, return_request_handler, req2, msg_size);
    }
  else
    {
      /*
       * Found closure.  Apply result to it.
       */
      Closure *t;
      
      parent->acquire_level_needed = 2;
      t = Closure_return_local(parent, req->child_num, &req->stuff);
      
      if (t == NULL)
	{
	  /*
	   * Nothing to provably-good steal.
	   */
	  /* Cilk_dprintf("sending pgs failure to machine %d\n", req->stealer.machine); */
	  send_steal_failure(req->stealer);
	}
      else
	{
	  /*
	   * Provably good steal!
	   */
	  
	  Closure_lock(t);
	  /* Cilk_dprintf("sending pgs success to machine %d\n", req->stealer.machine); */
	  CILK_ASSERT(t->status == CLOSURE_READY);
	  pack_up_and_send(t, req->stealer, USE_PRIVATE(Process));
	}
    }
}

/*
 * Return protocol.
 *
 * This protocol deposits the return value of a subcomputation (child)
 * into the appropriate place (parent).  This function returns a
 * closure to be executed next, or NULL if none.
 * The child must not be locked by ourselves, and be in no deque.
 */
static Closure *Closure_return(Closure *child)
{
     ClosureID parent;
     ClosureID forward;
     Closure *res;
     ReturnStuff *stuff;
     int child_num;

     CILK_ASSERT(child);
     CILK_ASSERT(child->join_counter == 0);
     CILK_ASSERT(child->status == CLOSURE_RETURNING);
     CILK_ASSERT(child->owner_ready_deque == NOBODY);
     Closure_assert_alienation(child);

     stuff = child->return_stuff;
     child_num = child->child_num;
     
     parent = child->parent;
     CILK_ASSERT(parent.c != NULL);

     /* All info has been extracted from the child - now the child
      * is no longer needed */
     /*
      * the lock is not really necessary, since if rule D
      * is obeyed no one can have a pointer to child at this
      * point.  But, better be paranoid...
      */
     WHEN_CILK_DEBUG(Closure_lock(child));
     Closure_destroy(child);
     WHEN_CILK_DEBUG(child = NULL);

     /* 
      * at this point the status is as follows: the child is
      * deallocated.  However, the child is still
      * in the children list of the parent.  Return "stuff"
      * to the parent.
      */
     
     if (parent.machine == USE_MACHINE(Machine))
       {
	 Closure_lock(parent.c);
	 
	 if (parent.c->status == CLOSURE_FORWARDING)
	   {
	     forward = parent.c->forward;
	     CILK_ASSERT(forward.machine >= 0);
	     parent.c->join_counter--;
	     if (parent.c->join_counter == 0)
	       Closure_destroy(parent.c);
	     else
	       Closure_unlock(parent.c);
	     goto forward;
	   }
	 
	 Cilk_memory_release(NULL, 0, 1);
	 
	 res = Closure_return_local(parent.c, child_num, stuff);
	 
	 /*
	  * If we get a provably good steal, do the appropriate
	  * acquire to get updated memory.
	  */
	 if (res)
	   {
	     CILK_ASSERT(res->status == CLOSURE_READY);
	     Cilk_memory_acquire(NULL, 0, res->acquire_level_needed);
	     res->acquire_level_needed = 1;
	   }
       }
     else
       {
	 /*
	  * Parent is remote.
	  */
	 ReturnRequest *req;
	 int msg_size;
	 
	 forward = parent;
	 CILK_ASSERT(forward.machine >= 0);
	 
       forward:
	 Cilk_memory_release(NULL, 0, 2);

	 /* Cilk_dprintf("remote return to %d\n", forward.machine); */
	 
	 /*
	  * Allocate a closure for a provably-good steal
	  */
	 res = Cilk_Closure_create();
	 res->status = CLOSURE_NOT_VALID_YET;
	 
	 /*
	  * Allocate and fill in return message
	  */
	 msg_size = sizeof(ReturnRequest) + ALIGN(stuff->value_size, CILK_MAX_ALIGN);
	 req = Cilk_alloc_msg_buf(msg_size, forward.machine);
	 req->child_num = child_num;
	 req->stealer = Closure_make_ID(res);
	 req->parent = forward.c;
	 req->stuff = *stuff;
	 memcpy(req->stuff.value, stuff->value, ALIGN(stuff->value_size, CILK_MAX_ALIGN));
	 
	 Cilk_rpc(forward.machine, return_request_handler, req, msg_size);
	 
	 while(res->status == CLOSURE_NOT_VALID_YET)
	   Cilk_poll();
	 
	 if (res->status == CLOSURE_STEAL_FAILURE)
	   {
	     /* Cilk_dprintf("received pgs failure\n"); */
	     Closure_destroy(res);
	     res = NULL;
	   }
	 else
	   {
	     /* Cilk_dprintf("received pgs success\n"); */
	     CILK_ASSERT(res->status == CLOSURE_READY);
	     /* CILK_ASSERT(res->magic == CILK_CLOSURE_MAGIC); */
	     CILK_ASSERT(res->acquire_level_needed == 2);
	     res->acquire_level_needed = 1;
	     
	     Cilk_memory_acquire(NULL, 0, 2);
	   }
       }
     
     Cilk_free_machine(stuff, sizeof(ReturnStuff) + ALIGN(stuff->value_size, CILK_MAX_ALIGN));
     
     return res;
}

/*
 * suspend protocol 
 */
static void Closure_suspend(Closure *cl)
{
     Closure *cl1;

     Closure_checkmagic(cl);
     Closure_assert_ownership(cl);
     deque_assert_ownership(USE_PRIVATE(Process));

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);
     cl->status = CLOSURE_SUSPENDED;

     Cilk_event(EVENT_SUSPEND);
     cl1 = deque_xtract_bottom(USE_PRIVATE(Process));
     CILK_ASSERT(cl == cl1);
}

/*************************************************************
 * Various interactions with the cilk2c output
 *************************************************************/

/* at a slow sync; return 0 if the sync succeeds, and 1 if suspended */
int Cilk_sync(void)
{
     Closure *t;
     int res = 0;

     Cilk_event(EVENT_CILK_SYNC);

     deque_lock(USE_PRIVATE(Process));

     t = deque_peek_bottom(USE_PRIVATE(Process));

     deque_unlock(USE_PRIVATE(Process));

     /* run all pending inlets */
     poll_inlets(t);

     Cilk_memory_release(NULL, 0, 1);
     
     deque_lock(USE_PRIVATE(Process));
     CILK_ASSERT(t == deque_peek_bottom(USE_PRIVATE(Process)));

     Closure_lock(t);
     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     /* CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC); */

     /* assert we are really at the top of the stack */
     CILK_ASSERT(CLOSURE_HEAD(t) + 1 == CLOSURE_TAIL(t));

     if (t->join_counter != 0) {
	  Closure_suspend(t);
	  res = 1;
     }

     Closure_unlock(t);

     deque_unlock(USE_PRIVATE(Process));

     poll_inlets(t);

     CILK_ASSERT(t->returned_values == NULL);

     /* If we're going to continue running, get updated versions of memory */
     if (!res)
       {
	 Cilk_memory_acquire(NULL, 0, t->acquire_level_needed);
	 t->acquire_level_needed = 1;
       }

     return res;
}

#if CILK_TIMING
/* update work and CP after a slow sync */
void Cilk_after_sync_slow_cp(Cilk_time * work, Cilk_time * cp)
{
     Closure *t;
     Cilk_time sync_work, sync_cp;

     deque_lock(USE_PRIVATE(Process));
     t = deque_peek_bottom(USE_PRIVATE(Process));

     Closure_lock(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     /* CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC); */

     /* assert we are really at the top of the stack */
     CILK_ASSERT(CLOSURE_HEAD(t) + 1 == CLOSURE_TAIL(t));

     sync_work = t->work;
     sync_cp = t->cp;
     t->cp = 0;
     t->work = 0;

     Closure_unlock(t);
     deque_unlock(USE_PRIVATE(Process));

     /* these references can fault - must be outside locks! KHR */
     *work += sync_work;
     if (sync_cp > *cp)
          *cp = sync_cp;
}
#endif

static void move_result_into_closure(Closure *t, void *resultp,
				     int size)
{
     ReturnStuff *stuff;
     Closure_assert_ownership(t);
     
     stuff = Cilk_malloc_machine(sizeof(ReturnStuff) + ALIGN(size, CILK_MAX_ALIGN));
     stuff->value_size = size;
     WHEN_CILK_TIMING({
       stuff->cp = USE_PRIVATE(cp_hack);
       stuff->work = USE_PRIVATE(work_hack);
     });
     if (size)
       memcpy (stuff->value, (void *)resultp, size);
     
     t->return_stuff = stuff;
}

/*
 * This is the exception handler.
 *
 * So far, exceptions are only caused by steals.  The handler returns
 * 1 if the worker must return to the runtime system, 0 otherwise 
 */
int Cilk_exception_handler(void *resultp, int size)
{
     Closure *t;
     int res = 0;

     Cilk_event(EVENT_EXCEPTION);

     deque_lock(USE_PRIVATE(Process));
     t = deque_peek_bottom(USE_PRIVATE(Process));

     CILK_ASSERT(t);

     Closure_lock(t);

     reset_exception_pointer(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING ||
		 t->status == CLOSURE_RETURNING);

     if (CLOSURE_HEAD(t) >= CLOSURE_TAIL(t)) {
	  res = 1;

	  if (t->status == CLOSURE_RUNNING) {
	       t->status = CLOSURE_RETURNING;
	       move_result_into_closure(t, resultp, size);
	  }
     } else {
          /* CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC); */
     }

     Closure_unlock(t);
     deque_unlock(USE_PRIVATE(Process));

     return res;
}

/*
 * before a return in the slow code.  This function marks the
 * closure as returning, so that it won't be stolen.
 */
void Cilk_set_result(void *resultp, int size)
{
     Closure *t;

     Cilk_event(EVENT_RETURN_SLOW);

     deque_lock(USE_PRIVATE(Process));
     t = deque_peek_bottom(USE_PRIVATE(Process));
     Closure_lock(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     /* CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC); */

     t->status = CLOSURE_RETURNING;
     t->frame = (StackFrame *) NULL;

     move_result_into_closure(t, resultp, size);

     Closure_unlock(t);
     deque_unlock(USE_PRIVATE(Process));
}

/*********************************************************
 * Execution
 *********************************************************/
static Closure *setup_for_execution(Closure *t)
{
     Closure_assert_ownership(t);

     /* CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC); */

     t->status = CLOSURE_RUNNING;
     t->cache = USE_PRIVATE(cache) = cache_get_addr(USE_PRIVATE(Process));

     CLOSURE_STACK(t)[0] = t->frame;	/* push the first frame on the stack */
     CLOSURE_HEAD(t) = (volatile StackFrame**)CLOSURE_STACK(t);
     CLOSURE_TAIL(t) = (volatile StackFrame**)CLOSURE_STACK(t)+1;
     reset_exception_pointer(t);

     return t;
}

static Closure *return_value(Closure *t)
{
     Closure *res = NULL;

     Cilk_enter_state(STATE_RETURNING);

     CILK_ASSERT(t->status == CLOSURE_RETURNING);

     res = Closure_return(t);

     Cilk_exit_state(STATE_RETURNING);
     return res;
}

/*
 * execute the closure. If the closure is returning, take care of the
 * result.
 */
static Closure *do_what_it_says(Closure *t)
{
     Closure *res = NULL;
     StackFrame *f;

     /* t must not be locked */
     Closure_assert_alienation(t);
     Closure_lock(t);

     switch (t->status) {
	 case CLOSURE_READY:
	      setup_for_execution(t);
	      f = t->frame;
	      CILK_ASSERT(f);
	      CILK_ASSERT(f->sig);

	      Closure_unlock(t);

	      CILK_ASSERT(f->magic == CILK_STACKFRAME_MAGIC);

	      poll_inlets(t);

	      /* 
	       * MUST unlock the closure before locking the queue 
	       * (rule A in file PROTOCOLS)
	       */
	      deque_lock(USE_PRIVATE(Process));
	      deque_add_bottom(t, USE_PRIVATE(Process));
	      deque_unlock(USE_PRIVATE(Process));

	      /* now execute it */
	      Cilk_enter_state(STATE_WORKING);
	      (get_proc_slow(f->sig)) (f);
	      Cilk_exit_state(STATE_WORKING);

	      break;

	 case CLOSURE_RETURNING:
	      /*
	       * the return protocol assumes t is not locked,
	       * and everybody will respect the fact that t is
	       * returning
	       */
	      Closure_unlock(t);
	      res = return_value(t);
	      break;

	 default:
	      Cilk_die("BUG in do_what_it_says(), t->status = %d\n", t->status);
	      break;
     }

     return res;
}

/*********************************************************
 * The scheduler itself
 *********************************************************/
void Cilk_scheduler(Closure *t)
{
     /* 
      * t contains 'the next thing to do'.  Initially, the
      * scheduler on proc 0 executes the main closure.
      */
     Cilk_enter_state(STATE_TOTAL);

     while (!USE_MACHINE(done)) {
	  if (!t) {
	       /* try to get work from our local queue */
	       deque_lock(USE_PRIVATE(Process));
	       t = deque_xtract_bottom(USE_PRIVATE(Process));
	       deque_unlock(USE_PRIVATE(Process));
	  }

	  while (!t && !USE_MACHINE(done)) {
	       /* otherwise, steal */
	       Cilk_poll();
	       Cilk_enter_state(STATE_STEALING);
	       t = Closure_steal();
	       if (!t && USE_PRIVATE(yield_slice)) {
		    Cilk_lower_priority();
	       }
	       Cilk_exit_state(STATE_STEALING);
	  }

	  if (USE_PRIVATE(yield_slice))
	       Cilk_raise_priority();

	  if (!USE_MACHINE(done))
	       t = do_what_it_says(t);

	  /* 
	   * if provably-good steals happened, t will contain
	   * the next closure to execute
	   */
     }

     /*
      * Get rid of last invoke_main closure, just to keep the
      * statistics clean.
      */
     deque_lock(USE_PRIVATE(Process));
     while(1)
       {
	 t = deque_xtract_bottom(USE_PRIVATE(Process));
	 if (!t) break;
	 Closure_destroy(t);
       }
     deque_unlock(USE_PRIVATE(Process));

     Cilk_exit_state(STATE_TOTAL);
     return;
}

/*
 * initialization of the scheduler. 
 */
void Cilk_scheduler_init_machine(const Options *options)
{
     CILK_CHECK(options->permach > 0,
		("Partition size must be positive\n"));
     init_deques(options);
     init_caches(options);
     Cilk_mutex_init(&USE_MACHINE(remote_steal_lock));
}

void Cilk_scheduler_init_local(const Options *options)
{
     CILK_ASSERT(Self >= 0);
     rts_srand(Self * 162347);
}

/*
 * Dummy abort code
 */
void Cilk_abort_slow(void)
{
}
void Cilk_abort_standalone(void)
{
}
void Cilk_exit(int ret)
{
  char c;
  
  fflush(NULL);
  
  c = 4;
  write(1, &c, 1);
  c = 5;
  write(2, &c, 1);
  while(1) Cilk_poll();        /* wait for kill from global process */
}
