/*
 * $Id: shared-malloc.c,v 1.1.2.3 1997/12/22 20:52:23 randall Exp $
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <strings.h>

/*
 * This code is functional, but it really sucks.  We need to invent
 * a good parallel distributed malloc.
 */

typedef struct header
{
  struct header *next;
  int bucket;
} header;

DECLARE_MACHINE_VARIABLE(static void *, free_ptr, );
DECLARE_MACHINE_VARIABLE(static void *, end_ptr, );

DECLARE_MACHINE_VARIABLE(void *, heap, );
DECLARE_MACHINE_VARIABLE(size_t, heap_size, );

#define EXTRA_SPACE 64

#define NBUCKETS 32

const unsigned bsize[32] = { 16, 32, 64, 128, 256, 1024, 2048,
			     (1<<12) + EXTRA_SPACE,
			     (1<<13) + EXTRA_SPACE,
			     (1<<14) + EXTRA_SPACE,
			     (1<<15) + EXTRA_SPACE,
			     (1<<16) + EXTRA_SPACE,
			     (1<<17) + EXTRA_SPACE,
			     (1<<18) + EXTRA_SPACE,
			     (1<<19) + EXTRA_SPACE,
			     (1<<20) + EXTRA_SPACE,
			     (1<<21) + EXTRA_SPACE,
			     (1<<22) + EXTRA_SPACE,
			     (1<<23) + EXTRA_SPACE,
			     (1<<24) + EXTRA_SPACE,
			     (1<<25) + EXTRA_SPACE,
			     (1<<26) + EXTRA_SPACE,
			     (1<<27) + EXTRA_SPACE,
			     (1<<28) + EXTRA_SPACE,
			     (1<<29) + EXTRA_SPACE,
			     (1<<30) + EXTRA_SPACE,
			     (1<<31) + EXTRA_SPACE,
			     -1};

DECLARE_MACHINE_VARIABLE(struct header **, buckets, );

DECLARE_MACHINE_VARIABLE(Cilk_mutex, malloc_lock, );

void *Cilk_malloc(size_t size)
{
  header *h;
  int i;
  
  for(i=0; 1; i++)
    if (size + sizeof(header) <= bsize[i]) break;
  
  Cilk_mutex_wait(&USE_MACHINE(malloc_lock));
  
  h = USE_MACHINE(buckets)[i];
  if (h != NULL)
    {
      USE_MACHINE(buckets)[i] = h->next;
      Cilk_mutex_signal(&USE_MACHINE(malloc_lock));
      return h+1;
    }
  
  h = USE_MACHINE(free_ptr);
  USE_MACHINE(free_ptr) = (void*)h + bsize[i];
  CILK_CHECK(USE_MACHINE(free_ptr) <= USE_MACHINE(end_ptr), 
	     ("out of malloc space! Ask for a larger heap with `-heap'. (%d)\n", USE_MACHINE(Machine)));
  
  Cilk_mutex_signal(&USE_MACHINE(malloc_lock));
  h->bucket = i;
  return h+1;
}

void *Cilk_malloc_fixed(size_t size)
{
  void *m;
  
  size = ALIGN(size, CILK_MAX_ALIGN);
  
  Cilk_mutex_wait(&USE_MACHINE(malloc_lock));
  
  m = USE_MACHINE(free_ptr);
  USE_MACHINE(free_ptr) = m + size;
  CILK_CHECK(USE_MACHINE(free_ptr) <= USE_MACHINE(end_ptr), 
	     ("out of malloc space! (%d)\n", USE_MACHINE(Machine)));
  
  Cilk_mutex_signal(&USE_MACHINE(malloc_lock));
  
  return m;
}

void *Cilk_calloc(size_t nelem, size_t elsize)
{
  void *m = Cilk_malloc(nelem*elsize);
  if (m != NULL)
    bzero(m, nelem*elsize);
  return m;
}

void *Cilk_memalign(size_t alignment, size_t size)
{
  /* hack for now!!! KHR */
  CILK_CHECK(0, ("memalign unimplemented!\n"));
  return NULL;
}

void Cilk_free(void *p)
{
  header *h = (header*)p - 1;
  int i = h->bucket;
  
  Cilk_mutex_wait(&USE_MACHINE(malloc_lock));
  
  h->next = USE_MACHINE(buckets)[i];
  USE_MACHINE(buckets)[i] = h;
  
  Cilk_mutex_signal(&USE_MACHINE(malloc_lock));
}

void Cilk_malloc_init_machine(const Options *options)
{
  USE_MACHINE(buckets) = Cilk_malloc_machine_fixed(NBUCKETS * sizeof(header*));
  
  /* Initialization - figure out local machine's chunk of the heap */
  CILK_ASSERT(USE_MACHINE(heap) ==
	      ALIGN(USE_MACHINE(heap), CILK_MAX_ALIGN));
  USE_MACHINE(free_ptr) = USE_MACHINE(heap) +
    ALIGN(USE_MACHINE(Machine) * USE_MACHINE(heap_size) / options->nmach, CILK_MAX_ALIGN);
  USE_MACHINE(end_ptr) = USE_MACHINE(heap) +
    ALIGN((USE_MACHINE(Machine)+1) * USE_MACHINE(heap_size) / options->nmach, CILK_MAX_ALIGN);
}

static void init_frame_allocator(const Options *options);

void Cilk_malloc_init_local(const Options *options)
{
  init_frame_allocator(options);
}

/***********************************************************\
 * Fast allocator for frames
\***********************************************************/

DECLARE_PRIVATE_VARIABLE(static void *, pool_begin, );
DECLARE_PRIVATE_VARIABLE(static void *, pool_end, );
DECLARE_PRIVATE_VARIABLE(void *, im_free_lists, [CILK_FRAME_FREELISTS]);

WHEN_CILK_DEBUG(DECLARE_PRIVATE_VARIABLE(struct Cilk_im_stats *, frame_info, );)

#define CILK_FRAME_CHUNK (32 * 1024)

void Cilk_malloc_done_local(const Options *options)
{
  /* WHEN_CILK_DEBUG(Cilk_dprintf("used:%d wasted:%d allocated:%d\n", USE_PRIVATE(frame_info)->used, USE_PRIVATE(frame_info)->wasted, USE_PRIVATE(frame_info)->allocated)); */
}

static void init_frame_allocator(const Options *options)
{
  int i;
  for (i = 0; i < CILK_FRAME_FREELISTS; ++i)
    USE_PRIVATE(im_free_lists)[i] = NULL;
  WHEN_CILK_DEBUG(USE_PRIVATE(frame_info) = malloc(sizeof(struct Cilk_im_stats)));
  WHEN_CILK_DEBUG(USE_PRIVATE(frame_info)->used = USE_PRIVATE(frame_info)->wasted = USE_PRIVATE(frame_info)->allocated = 0);
}

static void allocate_private_pool(void)
{
     USE_PRIVATE(pool_begin) = Cilk_malloc_fixed(CILK_FRAME_CHUNK);
     CILK_CHECK(USE_PRIVATE(pool_begin) != NULL,
		("frame malloc() out of memory\n"));
     USE_PRIVATE(pool_end) = USE_PRIVATE(pool_begin) +
	  CILK_FRAME_CHUNK;
     WHEN_CILK_DEBUG(USE_PRIVATE(frame_info)->allocated += CILK_FRAME_CHUNK);
}


/*
 * We only reach here if the fast version (in cilk.h) fails
 * because there aren't any frames of the correct size in
 * the free list.
 */
void *Cilk_internal_malloc(size_t size)
{
  void *mem;

  CILK_ASSERT(size <= CILK_INTERNAL_MALLOC_THRESHOLD);
  
  WHEN_CILK_DEBUG(USE_PRIVATE(frame_info)->used += size);
  
  size = Cilk_internal_malloc_canonicalize(size);
  
tryagain:
  mem = USE_PRIVATE(pool_begin);
  USE_PRIVATE(pool_begin) = mem + size;
  if (USE_PRIVATE(pool_begin) > USE_PRIVATE(pool_end))
    {
      WHEN_CILK_DEBUG(USE_PRIVATE(frame_info)->wasted += USE_PRIVATE(pool_end) - mem);
      allocate_private_pool();
      goto tryagain;
    }
  return mem;
}

/*
 * support for alloca
 */
#if CILK_ALLOCA
DECLARE_PRIVATE_VARIABLE(extern struct ClosureCache *const, cache ,);

void *Cilk_alloca(size_t size) 
{
     StackFrame *f = (StackFrame *)(USE_PRIVATE(cache)->tail)[-1];
     struct alloca_header *p;

     size += sizeof(struct alloca_header);
     p = Cilk_internal_malloc_fast(size);
     p->size = size;
     p->next = f->alloca_h;
     f->alloca_h = p;
     return (void *)(p+1);
}
#endif /* CILK_ALLOCA */

